/***************************************************** Flash Programmer  ****************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following code is used to program the flash on all Red Rapids products.

	Note:	The application can be applied to the maximum number of Red Rapids devices supported by the host.

	Rev History:  12-09-2019:	Initialize *p_FlashCode and *p_Record to NULL.
				  09-05-2018:	Initial release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*****************************************************  Flash Programmer  ****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include "driver_functions.h"					// Functions to communicate with the UIO and DMA drivers.
#include "flash.h"								// Flash programmer application.
#include "global_functions.h"					// Functions that apply to all products.
#include "programmer_functions.h"				// Functions used to program the flash device.

int main(int argc, char* argv[])
{
	s_UIOScan UIO[NUM_DEVICES];					// Descriptor for each UIO device identified as a Red Rapids product.
	s_Handle *p_HW[NUM_DEVICES];				// Hardware handle to each Red Rapids device.
	int status;									// Variable to store status returned by function calls.
	int index;									// Variable to index array elements in loops.
	char mcsfilename[64] = "No File Specified";	// Character string to store MCS file name.
	s_FlashCode *p_FlashCode = NULL;					// Pointer to the first record in the linked list of flash code.
	s_FlashCode *p_Record = NULL;						// Pointer to a flash code record.
	unsigned directive = 0;						// Directive requested by a command line argument.
	unsigned segment = 0;						// Segment of flash requested by a command line argument.
	unsigned devnum = 0;						// Device number of Red Rapids hardware to program.
	unsigned address_start;						// Starting address of a requested segment.
	unsigned address_end;						// Ending address of a requested segment.


	printf("***************************** Flash Programmer *****************************\n");

	//////////////////////////////////////  Parse Command Line Arguments And MCS File  ///////////////////////////////////////

	printf("\nChecking command inputs...\n");

	//Proceesing command line arguments.
	for(index = 1; index < argc; index++)
	{
		//Print the command usage to the console if requested with a "help" argument.
		if(!strcmp(argv[index],"-help") || !strcmp(argv[index],"-h") || !strcmp(argv[index],"-?") || !strcmp(argv[index],"?"))
		{
			printf("\nCOMMAND USAGE:\n");
			printf("\nflash [-p -g -a1 -a2] [-b -e -u <filename> -v <filename>] [-d <optional device number>]\n");
			printf("\nOne of the following four flash segments must be selected:\n");
			printf(" -p\tApply directive to the primary segment.\n");
			printf(" -g\tApply directive to the golden segment.\n");
			printf(" -a1\tApply directive to alternate segment #1.\n");
			printf(" -a2\tApply directive to alternate segment #2.\n");
			printf("\nOne of the following four directives must be selected:\n");
			printf(" -b\t\tBlank check the entire selected segment.\n");
			printf(" -e\t\tErase the entire selected segment.\n");
			printf(" -u <filename>\tUpdate the selected segment <MCS file required>.\n");
			printf(" -v <filename>\tVerify the selected segment <MCS file required>.\n");
			printf(" -d <optional device number>\tSelect a target device number other than zero.\n");
			exit(0);
		}
		// Check for more than one segment select switch.
		else if(segment && (!strcmp(argv[index],"-p") || !strcmp(argv[index],"-g") || !strcmp(argv[index],"-a1") || !strcmp(argv[index],"-a2")))
		{
			printf("\tERROR: Only one segment allowed (-p, -g, -a1 or -a2).\n");
			exit(1);
		}
		// Set the golden segment flag if the switch is selected.
		else if(!strcmp(argv[index],"-g"))
			segment = 0x4;
		// Set the primary segment flag if the switch is selected.
		else if(!strcmp(argv[index],"-p"))
			segment = 0x5;
		// Set the alternate segment A flag if the switch is selected.
		else if(!strcmp(argv[index],"-a1"))
			segment = 0x6;
		// Set the alternate segment B flag if the switch is selected.
		else if(!strcmp(argv[index],"-a2"))
			segment = 0x7;
		// Check for more than one directive switch.
		else if(directive && (!strcmp(argv[index],"-b") || !strcmp(argv[index],"-e") || !strcmp(argv[index],"-u") || !strcmp(argv[index],"-v")))
		{
			printf("\tERROR: Only one directive (-b, -e, -u, or -v) allowed.\n");
			exit(1);
		}
		// Set the blank check directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-b"))
			directive = 1;
		// Set the erase directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-e"))
			directive = 2;
		// Set the update directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-u"))
		{
			directive = 3;
			strcpy(mcsfilename, argv[++index]);
		}
		// Set the verify directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-v"))
		{
			directive = 4;
			strcpy(mcsfilename, argv[++index]);
		}
		// Set the target device number if the switch is selected.
		else if(!strcmp(argv[index],"-d"))
			devnum = atoi(argv[++index]);
		// Report an error if there is an invalid command line argument.
		else 
		{
			printf("\tERROR: Invalid argument entered on the command line (%s).\n", argv[index]);
			exit(1);
		}
	}

	// Check for illegal combinations of command line switches.
	if(!directive)
		{
			printf("\tERROR: At least one directive (-b, -e, -u, or -v) must  be selected.\n");
			exit(1);
		}
	if(!segment)
		{
			printf("\tERROR: At least one segment (-g, -p, -a1 or -a2) must  be selected.\n");
			exit(1);
		}
	if((directive == 3 || directive == 4) && !strcmp(mcsfilename,"No File Specified"))
		{
			printf("\tERROR: The update (-u) or verify (-v) directives require an MCS file.\n");
			exit(1);
		}

	// Parse the MCS file if it was selected.
	if (strcmp(mcsfilename,"No File Specified"))
	{
		// Add the .mcs file extension if it is not inclded in the file name.
		if (strstr(mcsfilename,".mcs") == NULL && strstr(mcsfilename,".MCS") == NULL)
			strcat(mcsfilename,".mcs");
		// Parse the MCS file to extract the flash code.
		p_FlashCode = MCSParser(mcsfilename);
		// Flag an error if a NULL pointer is returned.
		if (p_FlashCode == NULL)
		{
			printf("\tERROR: Parsing the MCS file (%s) failed.\n", mcsfilename);
			exit(1);
		}
	}

	///////////////////////////////////////////  Establish UIO Driver Communication  ///////////////////////////////////////////

	printf("\nScanning for Red Rapids hardware...\n");
	
	// Scan the UIO devices for Red Rapids devices.
	if (!UIOScan(UIO))
	{
		printf("\n...terminating execution because no Red Rapids hardware found.\n");
		return EXIT_FAILURE;
	}

	// Allocate memory for the structure and initialize if successful.
	if ((p_HW[devnum] = malloc(sizeof *p_HW[devnum])) == NULL)
	{
		printf("\tERROR: Unable to allocate memory for the device #%u hardware handle.\n", devnum);
		// Free any allocated memory.
		FreeResources(p_HW[devnum], UIO[devnum], p_FlashCode);
		printf("\n...terminating execution due to hardware handle memory allocation error.\n");
		return EXIT_FAILURE;
	}
	else
		// Initialize each member of the device handle structure.
		DEVInit(p_HW[devnum], devnum);

	//////////////////////////////////////////////////  Initialize Hardware  //////////////////////////////////////////////////
	
	printf("\nInitializing device DEV%u...\n", devnum);

	// Open the UIO character device file assigned to each Red Rapids device.
	if ((status = DEVOpen(UIO[devnum], &p_HW[devnum]->DEV)))
		printf("\t[ERROR] UIO character device could not be opened.\n");
	else
	{
		// Display the firmware revision date.
		printf("\t[INFO] Firmware revision (MMDDYYYY): %08X\n", FirmwareRevision(p_HW[devnum]));

		// Report the model number returned from the Red Rapids device.
		printf("\t[INFO] Identified Red Rapids Model %03X.\n", ModelNumber(p_HW[devnum]));

		// Software reset the device.
		SoftwareReset(p_HW[devnum]);
			
		// Pause 100 useconds for the programming clock to settle.
		MicrosecondTimer(100);

		// Configure the programmer if the clock is stable.
		if ((status = GetClockStatus(p_HW[devnum])))
			printf("\t[ERROR] Programming clock not stable.\n");
		else
		{
			printf("\t[INFO] Programming clock stable.\n");

			// Enable the flash chip.
			ProgEnable(p_HW[devnum]);

			// Set the selected segment.
			ProgSegment(p_HW[devnum], segment);

			// Initialize the flash chip.
			ProgInit(p_HW[devnum]);
			
			// Print the flash device ID to the screen.
			printf("\t[INFO] Flash Device ID: 0x%04X\n", DeviceID(p_HW[devnum]));
			
			// Compare the flash capacity to the selected segment.
			if (DeviceID(p_HW[devnum]) != 0x887E && segment > 0x5)
			{
				printf("\t[ERROR] Selected segment not available.\n");
				status = FLASH_ERROR;
			}
		}
	}

	// Exit if any errors were detected.
	if (status)
	{
		// Free any allocated memory.
		FreeResources(p_HW[devnum], UIO[devnum], p_FlashCode);
		printf("\n...terminating execution due to hardware initialization errors.\n");
		return EXIT_FAILURE;
	}
		
	////////////////////////////////////////////  Execute Programmer Directives  //////////////////////////////////////////////

	printf("\n[WARNING] The following operations may take several minutes to complete.");
	printf("\n[WARNING] Any interruption may require an external programmer to recover!\n");

	// Set the starting and ending address.
	if (directive == 3 || directive == 4)
	{
		// Find the last record in the flash code.
		p_Record = p_FlashCode;
		while (p_Record->next != NULL)
			p_Record = p_Record->next;

		// Set the address range equal to the flash code.
		address_start = p_FlashCode->address;
		address_end = p_Record->address;
	}
	else
	{
		// Set the address range equal to an entire segment.
		address_start = 0x0;
		address_end = (BLOCK_SIZE*BLOCK_COUNT/SEGMENT_COUNT) - BLOCK_SIZE;
	}
		
	// Perform blank check operation.
	if (directive == 1)
	{
		printf("\nPerforming blank check operation...\n");
		if (BlankCheck(p_HW[devnum], address_start, address_end))
			printf("\t[INFO] Segment is not blank.\n");
		else
			printf("\t[INFO] Segment is blank.\n");
	}

	// Perform erase operation.
	if (directive == 2 || directive == 3)
	{
		printf("\nPerforming unlock operation...\n");
		// Unlock the selected address range.
		if (Unlock(p_HW[devnum], address_start, address_end))
			printf("\t[ERROR] Segment unlock failed.\n");
		else
		{
			// Unlock completed without errors.
			printf("\t[INFO] Segment is unlocked.\n");

			printf("\nPerforming erase operation...\n");
			// Erase the selected address range.
			if (Erase(p_HW[devnum], address_start, address_end))
				printf("\t[ERROR] Erase operation failed.\n");
			else
			{
				// Erase completed without errors.
				printf("\t[INFO] Segment is erased.\n");
				
				// Perform program operation.
				if (directive == 3)
				{
					printf("\nPerforming program operation...\n");
					if (Program(p_HW[devnum], p_FlashCode))
						// Programming operation reported errors.
						printf("\t[ERROR] Segment program failed.\n");
					else
						// Programming completed without errors.
						printf("\t[INFO] Segment is programmed.\n");
				}
			}
		}
	}

	// Perform verify operation.
	if (directive == 4)
	{
		printf("\nPerforming verify operation...\n");
		if (Verify(p_HW[devnum], p_FlashCode))
			printf("\t[INFO] Segment does not match MCS file.\n");
		else
			printf("\t[INFO] Segment matches MCS file.\n");
	}

	// Disable the flash chip.
	ProgDisable(p_HW[devnum]);

	/////////////////////////////////////////////////  Free Allocated Memory /////////////////////////////////////////////////

	// Free any allocated memory.
	FreeResources(p_HW[devnum], UIO[devnum], p_FlashCode);

	printf("\n...all operations completed.\n");
	return EXIT_SUCCESS;
}
