/***************************************************  Programmer Functions  ***************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to program a flash device.

	Rev History:	11-28-2018:	Corrected progress indicator in Program() plus style and format changes.
	Rev History:	09-05-2018:	Initial release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

***************************************************  Programmer Functions  ***************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "driver_functions.h"
#include "flash.h"
#include "global_functions.h"
#include "programmer_functions.h"

// Convert a character hex string to an unsigned long integer type.
unsigned HexToUnsigned(char *HexNumber)
{
	unsigned number = 0;	// Variable that will be returned by the function.
    size_t digits = 0;		// Number of digits in the string representing the unsigned integer.
    size_t loop = 0;		// Loop counter used to parse the digits in the string.
	unsigned basemult = 1;	// Base value used to extract the number.

	// Determine the number of digits to process.
	digits = strlen(HexNumber);
	
	// Perform a hex conversion of each digit to produce the final value.
	for (loop = 0; loop < digits; loop++)
	{
		switch (HexNumber[digits-loop-1])
		{
			case '0': {number += 0 * basemult; break;}
			case '1': {number += 1 * basemult; break;}
			case '2': {number += 2 * basemult; break;}
			case '3': {number += 3 * basemult; break;}
			case '4': {number += 4 * basemult; break;}
			case '5': {number += 5 * basemult; break;}
			case '6': {number += 6 * basemult; break;}
			case '7': {number += 7 * basemult; break;}
			case '8': {number += 8 * basemult; break;}
			case '9': {number += 9 * basemult; break;}
			case 'a':
			case 'A': {number += 10 * basemult; break;}
			case 'b':
			case 'B': {number += 11 * basemult; break;}
			case 'c':
			case 'C': {number += 12 * basemult; break;}
			case 'd':
			case 'D': {number += 13 * basemult; break;}
			case 'e':
			case 'E': {number += 14 * basemult; break;}
			case 'f':
			case 'F': {number += 15 * basemult; break;}
			default: {break;}
		}
		basemult *= 16;
	}
	
	// Return the unsigned long integer result.
    return number;
}

// Perform a blank check on the requested address range.
// Return of zero indicates blank, non-zero indicates not blank.
int BlankCheck(s_Handle *p_HW, unsigned AddressStart, unsigned AddressEnd)
{
	unsigned address = 0;
	int status = 0;
	
	// Set the starting address.
	address = AddressStart;

	// Unlock all of the blocks in the address range specified.
	while (address <= AddressEnd && !status)
	{
		// Clear the status register assigned to the current partition.
		StatusClear(p_HW, (address/PARTITION_SIZE)*PARTITION_SIZE);

		// Write the address of the block to check.
		ProgAddress(p_HW, address);

		// Issue the sequence of commands that will check the block.
		ProgWrite(p_HW, BLANK_CHECK);
		ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the blank check to complete.
		status = StatusCheck(p_HW, (address/PARTITION_SIZE)*PARTITION_SIZE, TYP_BLNKCHK_TIME/4);

		// Increment to the next block address.
		address += BLOCK_SIZE;
	}
	
	return status;
}

// Report the device ID of the flash chip.
// Returns the device ID in hex format.
unsigned DeviceID(s_Handle *p_HW)
{
	unsigned value = 0;

	// Read the flash device ID.
	ReadID(p_HW, 0x0, 0x1, &value);

	return (unsigned)value & 0xFFFF;
}

// Erase the address range requested.
// Return of zero indicates success, non-zero indicates failure.
int Erase(s_Handle *p_HW, unsigned AddressStart, unsigned AddressEnd)
{
	unsigned address = 0;
	unsigned status = 0;
	unsigned block_total;
	unsigned block_count = 0;
	unsigned percent_complete = 0;
	
	// Calculate the total number of blocks to process for progress reporting.
	block_total = (AddressEnd - AddressStart)/BLOCK_SIZE;

	// Set the starting address.
	address = AddressStart;

	// Erase all of the blocks in the address range specified.
	while (address <= AddressEnd && !status)
	{
		// Clear the status register assigned to the current partition.
		StatusClear(p_HW, (address/PARTITION_SIZE)*PARTITION_SIZE);

		// Write the address of the block to erase.
		ProgAddress(p_HW, address);

		// Issue the sequence of commands that will erase the block.
		ProgWrite(p_HW, BLOCK_ERASE);
		ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the erase to complete.
		status = StatusCheck(p_HW, (address/PARTITION_SIZE)*PARTITION_SIZE, TYP_ERASE_TIME/4);

		// Increment to the next block address.
		address += BLOCK_SIZE;
		
		// Report the progress in 10% increments.
		Progress(block_total, &block_count, &percent_complete, 10);

	}
	
	printf("\n");	
	return status;
}

// Free all resources allocated to a device.
// Return of zero indicates success, non-zero indicates an error condition.
int FreeResources(s_Handle *p_HW, s_UIOScan UIO, s_FlashCode *p_FlashCode)
{
	s_FlashCode *p_record = NULL;
	int status = OK;
	
	// Close the Adapter device.
	if (p_HW->DEV.fd >= 0)
		status = DEVClose(UIO, &p_HW->DEV);
	
	// Free the hardware handle.
	free(p_HW);

	// Free the memory allocated to the flash program code.
	while (p_FlashCode != NULL)
	{
		// Save the location of the next record in the linked list.
		p_record = p_FlashCode->next;
		// Free the current record in the linked list.
		free(p_FlashCode);
		// Move to the next record in the linked list.
		p_FlashCode = p_record;
	}
	
	return status;
}

// Report the status of the programmer clock.
// Return of zero indicates clock is stable.
unsigned GetClockStatus(s_Handle *p_HW)
{
	unsigned long long value = 0;
	
	// Save and clear the contents of the error status register by writing to the error status address.
	// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
	p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
	// Read the saved copy of the error status to determine if all converter clocks are locked.
	value = p_HW->DEV.BAR0[ERROR_STATUS];

	// The flash programmer clock error status is assigned to bit #7 on Model 2xx products with firmware released after 2017.
	if (((ModelNumber(p_HW) & 0xF00ull) == 0x200) && ((FirmwareRevision(p_HW) & 0xFFFFull) > 0x2017) && (value & 0x80ull))
		// The flash programmer clock is not locked.
		return CLOCK_LOCK;
	// The flash programmer clock error status is assigned to bit #5 on all other products.
	else if (value & 0x20ull)
		// The flash programmer clock is not locked.
		return CLOCK_LOCK;

	return OK;
}

// Count the number of records retrieved from the MCS file.
// Return the number of records.
unsigned MCSCount(s_FlashCode *p_FlashCode)
{
	s_FlashCode *p_record = NULL;
	unsigned recordcnt = 0;
	
	// Point to the first record in the linked list.
	p_record = p_FlashCode;
	
	// Count the number of MCS records.
	while (p_record->next != NULL)
	{
		recordcnt++;
		p_record = p_record->next;
	}
	
	return recordcnt;
}
	
// Create a linked list of programming records consisting of address and 16-bit data word.
// Return NULL if an error is encountered.
s_FlashCode* MCSParser(char *MCSFileName)
{
	
	#define MAXLINE 522  // Maximum number of characters in a single line of the file.

	s_FlashCode *p_record = NULL;
	s_FlashCode *p_record_start = NULL;
	s_FlashCode *p_record_previous = NULL;
	char input_line[MAXLINE];
	char byte_count[3];
	char address[9];
	char address_extended[5];
	char data[5];
	FILE *mcsfile;
	unsigned end_of_file = 0;
    unsigned index1 = 0;
	
	// Open the MCS file for processing.
	mcsfile = fopen(MCSFileName,"r");

	// Only proceed if the file was opened successfully.
	if (mcsfile != NULL)
	{
		// Keep parsing lines of text until the end of file record is detected.
		while (!end_of_file)
		{
			// Get the next line of text.
			if (fgets(input_line,MAXLINE,mcsfile) != NULL)
			{
				// End of file record type.			
				if (input_line[8] == '1')
					// Set the end of file flag to exit the function.			
					end_of_file = 1;
				else
				{
					// Extended linear address record type.			
					if (input_line[8] == '4')
					{
						for (index1 = 0; index1 < 4; index1++)
							address_extended[index1] = input_line[9+index1];
						address_extended[4] = '\0';
					}
					else
					{
						// Data record type.			
						if (input_line[8] == '0')
						{
							// Extract the byte count.
							byte_count[0] = input_line[1];
							byte_count[1] = input_line[2];
							byte_count[2] = '\0';
					
							// Assemble the complete address.
							for (index1 = 0; index1 < 4; index1++)
								address[index1] = address_extended[index1];
							for (index1 = 4; index1 < 8; index1++)
								address[index1] = input_line[3+index1-4];
							address[8] = '\0';

							// Strip 16-bit data words from the available bytes.
							for (index1 = 0; index1 < HexToUnsigned(byte_count)/2; index1++)
							{
								// Create a new member of the linked list.
								p_record = (s_FlashCode*) malloc(sizeof(s_FlashCode));

								// Set the next pointer null since this is now the last record in the list.
								p_record->next = NULL;
								// Determine whether this is the first pointer in the linked list.
								if (p_record_start == NULL)
									// Assign the address of the first record to the start pointer.
									p_record_start = p_record;
								else
									// Assign the current address to the next pointer of the previous record.
									p_record_previous->next = p_record;
								// Store the current pointer in case another record is added to the linked list.
								p_record_previous = p_record;
						
								// Extract the current 16-bit data word.
								// Note: The data is reordered to place the least significant byte on the right.
								data[0] = input_line[9+(4*index1)+2];
								data[1] = input_line[9+(4*index1)+3];
								data[2] = input_line[9+(4*index1)];
								data[3] = input_line[9+(4*index1)+1];
								data[4] = '\0';
					
								// Store the address and data word in the current record as unsigned long values.
								p_record->address = HexToUnsigned(address)/2 + index1;
								p_record->data = HexToUnsigned(data);
							}
						}
					}
				}
			}
			// The end of file was detected before the end of file record was processed.
			else
			{
				// Set the end of file flag to exit the function.
				end_of_file = 1;
				// Free the the allocated memory.
				while (p_record_start != NULL)
				{
					// Save the location of the next record in the linked list.
					p_record = p_record_start->next;
					// Free the current record in the linked list.
					free(p_record_start);
					// Move to the next record in the linked list.
					p_record_start = p_record;
				}
			}
		}
	}

	return p_record_start;
}

// Program the address range requested.
// Return of zero indicates success, non-zero indicates failure.
int Program(s_Handle *p_HW, s_FlashCode *p_FlashCode)
{
	unsigned address_last = 0;
	unsigned address_base = 0;
	unsigned address_block = 0;
	unsigned address_partition = 0;
	unsigned status = 0;
	s_FlashCode *p_record = NULL;
	unsigned index1 = 0;
	unsigned index2 = 0;
	unsigned record_total;
	unsigned record_count = 0;
	unsigned percent_complete = 0;
	
	// Get the total number of MCS records to process for progress reporting.
	record_total = MCSCount(p_FlashCode);

	// Point to the first record in the linked list.
	p_record = p_FlashCode;

	while (p_record != NULL && !status)
	{
		// Store the base address of the current buffer.
		address_base = p_record->address;
		// Calculate the partition address of the current buffer.
		address_partition = (address_base/PARTITION_SIZE)*PARTITION_SIZE;
		// Calculate the block address of the current buffer.
		address_block = (address_base/BLOCK_SIZE)*BLOCK_SIZE;

		// Clear the status register assigned to the current partition
		StatusClear(p_HW, address_partition);

		// Issue the buffered program command at the base address.
		ProgAddress(p_HW, address_base);
		ProgWrite(p_HW, BUFFERED_PROGRAM);

		// Load the number of words (minus one) at the block address.
		ProgAddress(p_HW, address_block);
		ProgWrite(p_HW, (BUFFER_SIZE-1));

		// Write an entire buffer of data in bursts that will not overflow the command FIFO.
		// Note: There are two commands (address + data) written for each program record.
		for (index1 = 0; index1 < BUFFER_SIZE/(CMD_FIFO_BURST/2); index1++)
		{
			// Wait until the command FIFO has enough room to write a complete burst.
			ProgWait(p_HW);
			// Write the next burst to the command FIFO.
			// Note: There are two commands (address + data) written for each program record.
			for (index2 = 0; index2 < CMD_FIFO_BURST/2; index2++)
				// Look for the last record in the linked list.
				if (p_record != NULL)
				{
					// Load the current address and data value.
					ProgAddress(p_HW, p_record->address);
					ProgWrite(p_HW, p_record->data);
					// Store the current address in case this is the end of the linked list.
					address_last = p_record->address;
					// Transition to the next record in the linked list.
					p_record = p_record->next;
					
					// Report the progress in 10% increments.
					Progress(record_total, &record_count, &percent_complete,  10);
				}
				// Pad the remainder of the buffer with 0xFFFF if there is no more data to program.
				else
				{
					// Fill the end of the last buffer with pad data. 
					ProgAddress(p_HW, ++address_last);
					ProgWrite(p_HW, 0xFFFF);
				}
		}

		// Issue the confirm command.
		ProgAddress(p_HW, address_base);
		ProgWrite(p_HW, CONFIRM_COMMAND);

		// Wait for the programming to complete.
		status = StatusCheck(p_HW, address_partition, TYP_BUFPROG_TIME/4);
	}
	
	printf("\n");	
	return status ;
}

// Report progress toward processing a job as a percentage of the total task.
void Progress(unsigned JobSize, unsigned *JobCounter, unsigned *PercentComplete, unsigned Percentage)
{
	// Count the number of records processed.
	*JobCounter += 1;
	
	// Monitor progress toward completing the next percentage increment.
	if (*JobCounter >= *PercentComplete * JobSize / 100) 
	{
		// Start a progress monitor display.
		if (!*PercentComplete)
		{
			printf("\tProgress:   0%%");
			fflush(stdout);
		}
		else
		{
			// Report another percentage increment complete.
			printf("\b\b\b\b%3u%%", *PercentComplete);
			fflush(stdout);
		}
		
		// Increment the percentage complete.
		*PercentComplete += Percentage;
	}
}
	
// Load an address into the programmer.
void ProgAddress(s_Handle *p_HW, unsigned Address)
{
	p_HW->DEV.BAR0[BPI_BUS] = ADDRESS_COMMAND | (unsigned long long)Address;
}

// Disable programming to the flash.
void ProgDisable(s_Handle *p_HW)
{
	// Disable the program voltage and chip select to the flash.
	p_HW->DEV.BAR0[BPI_BUS] = ENABLE_COMMAND;
}

// Enable the programming to the flash.
void ProgEnable(s_Handle *p_HW)
{
	// Enable the program voltage and chip select to the flash.
	p_HW->DEV.BAR0[BPI_BUS] = ENABLE_COMMAND | 0x1ull;
	
	// Pause 1 microsecond to stabilize.
	MicrosecondTimer(1);
}

// Initialize the segment to asynchronous read mode.
// Return of zero indicates success, non-zero indicates an error condition.
void ProgInit(s_Handle *p_HW)
{
	unsigned index = 0;

	// Set all of the partitions in the segment to asynchronous read mode.
	for (index = 0; index < PARTITION_COUNT/SEGMENT_COUNT; index++)
	{
		// Issue the program read configuration register at every partition address.
		ProgAddress(p_HW, (index*PARTITION_SIZE) | 0xD007);
		ProgWrite(p_HW, PROG_READ_CONFIG_REG1);
		ProgWrite(p_HW, PROG_READ_CONFIG_REG2);
	}
}

// Perform a read operation from the programmer.
void ProgRead(s_Handle *p_HW, unsigned *p_Value)
{
	unsigned long long value = 0;

	// Issue a command to read from the flash.
	p_HW->DEV.BAR0[BPI_BUS] = READ_COMMAND;

	// Wait for a reply from the programmer.
	do
	{
		// Pause 1 usecond to reduce bus traffic.
		MicrosecondTimer(1);

		// Read the reply valid status.
		value = p_HW->DEV.BAR0[BPI_BUS];
	}
	while (!(value & 0x0000000100000000ull));

	// Return the lower 32-bit word.
	*p_Value = (unsigned)(value & 0x00000000FFFFFFFFull);

	// Pop the value just read off the reply FIFO.
	p_HW->DEV.BAR0[BPI_IF_CTL] = 0x0ull;
}

// Load the revision select into the programmer.
void ProgSegment(s_Handle *p_HW, unsigned Revision)
{
	p_HW->DEV.BAR0[BPI_BUS] = SEGMENT_COMMAND | (unsigned long long)Revision;
}

// Wait until the command FIFO can accept another burst of data.
void ProgWait(s_Handle *p_HW)
{
	unsigned long long value = 0;
	unsigned ready = 0;

	// Wait for the command FIFO full flag to deassert.
	// Note: Full is programmed to allow space for a complete burst (CMD_FIFO_BURST).
	do
	{
		// Pause 5 usecond to reduce bus traffic.
		MicrosecondTimer(CMD_FIFO_BURST/4);

		// Read the command FIFO full status.
		value = p_HW->DEV.BAR0[BPI_IF_CTL];

		// The status bit conveys active high ready with firmware released after 2017.
		if (((FirmwareRevision(p_HW) & 0xFFFF) > 0x2017) && (value & 0x1ull))
			// The FIFO can accept another burst.
			ready++;
		// The status bit conveys active high full with firmware released before 2018.
		else if (!(value & 0x1ull))
			// The FIFO can accept another burst.
			ready++;

	}
	while (!ready);
}

// Perform a write operation to the programmer.
void ProgWrite(s_Handle *p_HW, unsigned Command)
{
	p_HW->DEV.BAR0[BPI_BUS] = WRITE_COMMAND | (unsigned long long)Command;
}

// Issue a command to read the ID register from the partition at the selected address.
void ReadID(s_Handle *p_HW, unsigned Address, unsigned Offset, unsigned *p_Value)
{
	// Issue the read ID command at the partition address.
	ProgAddress(p_HW, Address);
	ProgWrite(p_HW, READ_ID);

	// Set the offset of the desired register.
	ProgAddress(p_HW, Address | Offset);

	// Read the reply from the programmer.
	ProgRead(p_HW, p_Value);
}

// Check the partition status bit and only return when it indicates ready.
// Return of zero indicates success, non-zero indicates unexpected flash state.
int StatusCheck(s_Handle *p_HW, unsigned Address, unsigned Timer)
{
	unsigned value = 0;

	// Wait for the status to indicate that the partition is ready.
	do
	{
		// Pause to reduce bus traffic.
		MicrosecondTimer(Timer);
		// Read the status register assigned to the current partition.
		StatusRead(p_HW, Address, &value);
	}
	// Monitor the device ready bit.
	while (!(value & 0x00000080));

	// Check the status for unexpected results.
	if (value & 0xFF7F)
		return value & 0xFF7F;
	else
		return OK;
}

// Issue a command to clear the status register from the partition at the selected address.
void StatusClear(s_Handle *p_HW, unsigned Address)
{
	// Set the partition address.
	ProgAddress(p_HW, Address);

	// Issue the command to clear the register.
	ProgWrite(p_HW, CLEAR_STATUS_REGISTER);
}

// Issue a command to read the status register from the partition at the selected address.
void StatusRead(s_Handle *p_HW, unsigned Address, unsigned *p_Value)
{
	// Set the partition address.
	ProgAddress(p_HW, Address);

	// Read the reply from the programmer.
	ProgRead(p_HW, p_Value);
}

// Unlock the address range requested.
// Return of zero indicates success, non-zero indicates an error condition.
int Unlock(s_Handle *p_HW, unsigned AddressStart, unsigned AddressEnd)
{
	unsigned address = 0;
	unsigned value = 0;
	
	// Set the starting address.
	address = AddressStart;

	// Unlock all of the blocks in the address range specified.
	while (address <= AddressEnd)
	{
		// Write the address of the block to unlock.
		ProgAddress(p_HW, address);

		// Issue the sequence of commands that will unlock the block.
		ProgWrite(p_HW, UNLOCK_BLOCK);
		ProgWrite(p_HW, CONFIRM_COMMAND);

		// Verify that the block is unlocked.
		ReadID(p_HW, (address/PARTITION_SIZE)*PARTITION_SIZE, 0x2, &value);
		// Report an error if the block is marked locked.
		if (value & 0x00000001)
			return FLASH_ERROR;

		// Increment to the next address.
		address += BLOCK_SIZE;
	}
	
	return OK;
}

// Compare the MCS file to the code loaded in flash.
// Return of zero indicates a match, non-zero indicates a mismatch.
int Verify(s_Handle *p_HW, s_FlashCode *p_FlashCode)
{
	unsigned value = 0;
	s_FlashCode *p_record = NULL;
	unsigned record_total;
	unsigned record_count = 0;
	unsigned percent_complete = 0;
	
	// Get the total number of MCS records to process for progress reporting.
	record_total = MCSCount(p_FlashCode);

	// Point to the first record in the linked list.
	p_record = p_FlashCode;
	
	// Process the entire linked list of flash code.
	while (p_record->next != NULL)
	{
		// Issue the read array command at the code address.
		ProgAddress(p_HW, p_record->address);
		ProgWrite(p_HW, READ_ARRAY);

		// Compare the MCS data to the value returned by the flash.
		ProgRead(p_HW, &value);
		if (p_record->data != value)
		{
			printf("\n");
			return 1;
		}
		p_record = p_record->next;
		
		// Report the progress in 1% increments.
		Progress(record_total, &record_count, &percent_complete,  1);
	}
	printf("\n");
	
	return OK;
}