/*****************************************************  Chip Functions  ******************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to initialize specific chips used in the SigStream design.

	Rev History:  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*****************************************************  Chip Functions  ******************************************************/

#include <stdlib.h>

#include "driver_functions.h"
#include "sigstream.h"
#include "global_functions.h"
#include "chip_functions.h"


// Set the ADA4927 amplifier enables.
void ADA4927Write(s_Handle *p_HW, unsigned ChipSelect, unsigned PortNumber, unsigned Data)
{
	// Performa a read-modify-write operation to the command status port expander output settings register (0x5).
	if (Data)
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) | (0x1 << PortNumber));
	else
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) & (0xFEFF >> (8 - PortNumber)));
}

// Perform a serial write transaction to the AD5628 ADC chip.
void AD5628Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Command, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 20, NULL);

	// Write data to the AD5628 register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, (Command << 4) | Address, AD5628_INST, Data << 8, AD5628_DATA, NULL);
}

// Set the AD8000 amplifier enables.
void AD8000Write(s_Handle *p_HW, unsigned ChipSelect, unsigned PortNumber, unsigned Data)
{
	// Performa a read-modify-write operation to the command status port expander output settings register (0x5).
	if (Data)
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) | (0x1 << PortNumber));
	else
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) & (0xFEFF >> (8-PortNumber)));
}

// Perform a serial read transaction to the AD9142A DAC chip.
// Return the value read back from the clock distribution chip.
unsigned AD9142ARead(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the AD9142A register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, (0x1 << (AD9142A_INST-1)) | Address, AD9142A_INST, 0x0, AD9142A_DATA, &value);

	return (unsigned)value;
}

// Perform a serial write transaction to the AD9142A DAC chip.
void AD9142AWrite(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the AD9142A register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, Address, AD9142A_INST, Data, AD9142A_DATA, NULL);
}

// Perform a serial read transaction to the AD9512 clock distribution chip.
// Return the value read back from the clock distribution chip.
unsigned AD9512Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the AD9512 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, (0x1 << (AD9512_INST-1)) | Address, AD9512_INST, 0x0, AD9512_DATA, &value);

	return (unsigned)value;
}

// Perform a serial write transaction to the AD9512 clock distribution chip.
void AD9512Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the AD9512 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, Address, AD9512_INST, Data, AD9512_DATA, NULL);
}

// Initiate a fast start-up calibration cycle on the AD9652 ADC.
void AD9652Cal(s_Handle *p_HW, unsigned ChipSelect)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Send a background calibration reset command to the calibration configuration register.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, 0x4FB, AD9652_INST, 0x0C, AD9652_DATA, NULL);

	// Clear the background calibration reset command.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, 0x4FB, AD9652_INST, 0x08, AD9652_DATA, NULL);

	// Wait 1.3 seconds for the fast start-up calibration to complete (@ 310 Msps).
	MicrosecondTimer(1300000);
}

// Perform a serial read transaction to the AD9652 ADC chip.
// Return the value read back from the clock distribution chip.
unsigned AD9652Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the AD9652 register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, (0x1 << (AD9652_INST-1)) | Address, AD9652_INST, 0x0, AD9652_DATA, &value);

	return (unsigned)value;
}

// Perform a serial write transaction to the AD9652 ADC chip.
void AD9652Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the AD9652 register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, Address, AD9652_INST, Data, AD9652_DATA, NULL);
}

// Perform a serial read transaction to the AD9653 ADC chip.
// Return the value read back from the clock distribution chip.
unsigned AD9653Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the AD9653 register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, (0x1 << (AD9653_INST - 1)) | Address, AD9653_INST, 0x0, AD9652_DATA, &value);

	return (unsigned)value;
}

// Perform a serial write transaction to the AD9653 ADC chip.
void AD9653Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the AD9653 register at the requested address.
	Serial(p_HW, SPI_SDIO, RISING_EDGE, Address, AD9653_INST, Data, AD9653_DATA, NULL);
}

// Initiate a calibration cycle on the ADC12D1600 ADC.
void ADC12D1600Cal(s_Handle *p_HW, unsigned ChipSelect)
{
	// Enable calibration by performing a read-modify-write operation to the configuration register (0x0).
	ADC12D1600Write(p_HW, ChipSelect, 0x0, ADC12D1600Read(p_HW, ChipSelect, 0x0) | 0x8000);

	// Wait 50 mseconds for the calibration to complete.
	MicrosecondTimer(50000);

	// Disable calibration by performing a read-modify-write operation to the configuration register (0x0).
	ADC12D1600Write(p_HW, ChipSelect, 0x0, ADC12D1600Read(p_HW, ChipSelect, 0x0) & 0x7FFF);
}

// Perform a serial read transaction to the ADC12D1600 ADC.
// Return the value read back from the chip.
unsigned ADC12D1600Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the ADC12D1600 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, (0x1 << (ADC12D1600_INST-1)) | 0x40 | (Address << 1), ADC12D1600_INST, 0x0, ADC12D1600_DATA, &value);

	return value;
}

// Perform a serial write transaction to the ADC12D1600 ADC chip.
void ADC12D1600Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the ADC12D1600 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, 0x40 | (Address << 1), ADC12D1600_INST, Data, ADC12D1600_DATA, NULL);
}

// Set the ADL5566 amplifier enables.
void ADL5566Write(s_Handle *p_HW, unsigned ChipSelect, unsigned PortNumber, unsigned Data)
{
	// Performa a read-modify-write operation to the command status port expander output settings register (0x5).
	if (Data)
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) | (0x1 << PortNumber));
	else
		FXL6408Write(p_HW, ChipSelect, 0x5, FXL6408Read(p_HW, ChipSelect, 0x5) & (0xFEFF >> (8-PortNumber)));
}

// Perform a serial read transaction to the ADS42LB69 ADC chip.
// Return the value read back from the clock distribution chip.
unsigned ADS42LB69Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, &value);

	// Read the ADS42LB69 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, (0x1 << (ADS42LB69_INST-1)) | Address, ADS42LB69_INST, 0x0, ADS42LB69_DATA, &value);

	return (unsigned)value;
}

// Perform a serial write transaction to the ADS42LB69 ADC chip.
void ADS42LB69Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the ADS42LB69 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, Address, ADS42LB69_INST, Data, ADS42LB69_DATA, NULL);
}

// Perform a serial read transaction to the FXL6408 command/status port expander chip.
// Return the value read back from the chip.
unsigned FXL6408Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Read the FXL6408 register at the requested address.
	Serial(p_HW, I2C, I2C_READ, (ChipSelect << 16) | (Address << 8) | ChipSelect | 0x1, 0x18, 0x0, FXL6408_DATA, &value);

	return value;
}

// Perform a serial write transaction to the FXL6408 command/status port expander chip.
void FXL6408Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Write data to the FXL6408 register at the requested address.
	Serial(p_HW, I2C, I2C_WRITE, (ChipSelect << 8) | Address, 0x10, Data, FXL6408_DATA, NULL);
}

// Perform a serial read transaction to the INA3221 voltage monitor.
// Return the value read back from the chip.
unsigned INA3221Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Read the INA3221 register at the requested address.
	Serial(p_HW, I2C, I2C_READ, (ChipSelect << 16) | (Address << 8) | ChipSelect | 0x1, 0x18, 0x0, INA3221_DATA, &value);

	return value;
}

// Perform a serial write transaction to the INA3221 voltage monitor.
void INA3221Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Write data to the INA3221 register at the requested address.
	Serial(p_HW, I2C, I2C_WRITE, (ChipSelect << 8) | Address, 0x10, Data, INA3221_DATA, NULL);
}

// Generate an external synchronization pulse on the LMK01000 clock distribution chip SYNC pin.
void LMK01000Sync(s_Handle *p_HW, unsigned ChipSelect)
{
	// Enable the chip select that is attached to the SYNC pin.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Issue a phantom write transaction to enable the chip select.
	Serial(p_HW, SPI, RISING_EDGE, 0x0, 0x8, 0x0, 0x8, NULL);

	// Disable the chip select that is attached to the SYNC pin.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ALL_DESELECT, 16, NULL);
}

// Perform a serial write transaction to the LMK01000 clock distribution chip.
void LMK01000Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the LMK01000 register at the requested address.
	// Note: The Microwire interface transposes address and data.
	Serial(p_HW, MICROWIRE, RISING_EDGE, Data, LMK01000_DATA, Address, LMK01000_ADD, NULL);
}

// Perform a serial write transaction to the LTC1661 DAC chip.
void LTC1661Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Chip select the requested device on the SPI bus.
	Serial(p_HW, SPI_CSB, RISING_EDGE, 0x0, 0, ChipSelect, 16, NULL);

	// Write data to the LTC1661 register at the requested address.
	Serial(p_HW, SPI, RISING_EDGE, Address, LTC1661_INST, Data << 2, LTC1661_DATA, NULL);
}

// Perform a serial read transaction to the LTC2991 voltage/current/temperature monitor.
// Return the value read back from the chip.
unsigned LTC2991Read(s_Handle *p_HW, unsigned ChipSelect, unsigned Address)
{
	unsigned value = 0;

	// Read the LTC2991 register at the requested address.
	Serial(p_HW, I2C, I2C_READ, (ChipSelect << 16) | (Address << 8) | ChipSelect | 0x1, 0x18, 0x0, LTC2991_DATA, &value);

	return value;
}

// Perform a serial write transaction to the LTC2991 voltage/current/temperature monitor.
void LTC2991Write(s_Handle *p_HW, unsigned ChipSelect, unsigned Address, unsigned Data)
{
	// Write data to the LTC2991 register at the requested address.
	Serial(p_HW, I2C, I2C_WRITE, (ChipSelect << 8) | Address, 0x10, Data, LTC2991_DATA, NULL);
}