/****************************************************  Operating Defaults  *****************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following defaults are assigned to structures that configure operating variables.

	Rev History:  03-11-2019:	Added error interrupt mask defaults.
				  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Operating Defaults  *****************************************************/

#include "driver_functions.h"
#include "sigstream.h"
#include "global_functions.h"
#include "channel_functions.h"

#if defined(__cplusplus)
	extern "C" {
#endif

//////////////////////////////////////////////  Configure Receiver (RX) Channels  //////////////////////////////////////////////

// Initialize a default structure for the RX channels.
s_Channel RXDefault = {
	{ // SwitchSelect
		0,	// ChannelSelect				Select the receiver (RX) input for this Signal Stream (SS) channel. - Range: 0 (product default), 1 (RX#1), 2 (RX#2), 3 (RX#3), 4 (RX#4)
											// Product default value: SS#1 = RX#1, SS#2 = RX#2, SS#3 = RX#3, SS#4 = RX#4
		0,	// ChannelBypass				Bypass the receiver (RX) input and route a fixed data pattern to this Signal Stream channel. - Range: 0 (no bypass), 1 (constant), 2 (binary counter)
		0	// Amplitude					Maximum amplitude of synthetic data source. (e.g. constant) - Range: 0 to 2^32 - 1
	},
	{ // PayloadFormat
		0,	// PayloadPacking				Pack data in the DMA buffer as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		16,	// DataItemSize					Size of a data item in bits as defined by VITA 49. - Range: 8, 10, 12, 14, 16, 18, 20, 32
		0,	// EventTagOrder				Order of the over-range (ovr) and sync event tags in the item packing field. - Range: 0 (ovr, trigger) to 1 (trigger, ovr)
		0,	// EventTagMask					Mask applied to the over-range (ovr) and trigger event tag bits. - Range: 0 (blank all), 1 (blank ovr), 2 (blank trigger), 3 (blank none)
		0,	// EventTagIndex				Bit position of the event tag in the item packing field referenced to the LSB. - Range: 0, 2, 4, 6, 8, 10, 12, 14
		16,	// ItemPackingFieldSize			Size of the item packing field as defined by VITA 49 [ItemPackingFieldSize >= DataItemSize]. - Range: 8, 10, 12, 14, 16, 20*, 24*, 32* (*Not available for Model 276.)
		0	// ChannelTag					Optional channel tag as defined by VITA 49, - Range: 0 to 2^16 - 1
	},
	{ // DatapathControl
		{ // Transform
			0,	// Bandwidth Select			Optional filter/resampler to produce output bandwidths below Fs/2 using FIR filters.
											// Programmable coefficient FIR filter - Range: 0 (BW=Fs/2 real), 1 (BW=Fs/2 complex), 2 (BW set by programmable filters)
											// Fixed coefficient HBF FIR filter. - Range: 0 (BW=Fs/2 real), 1 (BW=Fs/2 complex), 2 (BW=Fs/4 complex), 3 (BW=Fs/8 complex) ... 11 (BW=Fs/2048 complex)
			1,	// Real Resampler			Decimation factor (input : output) achieved by throwing away samples of real data only, not complex data. - Range: 1 to 2^16 - 1
			0,	// Frequency Translation	Normalized down converter center frequency (2^23)*(Fc/Fs), applicable only with non-zero Filter Select (above) - Range: 0 to 2^22.
			0,	// Phase Offset				Starting phase offset expressed in increments of pi/2^22, valid only with non-zero Frequency Translation (above) - Range: 0 to 2^23 - 1.
			0,	// Reserved
			0	// Reserved
		},
		{ // Residue
			0,	// Channel Stop				Disposition of data residue trapped in the pipeline when the channel is deactivated. - Range: 0 to 7
			0,	// Channel Cycle			Disposition of data residue trapped in the pipeline between cycles. - Range: 0 to 7
		},
		{ // SyncInitMask
			1,	// Binary Data Counter		Send a synchronization strobe to the binary data counter immediately before processing the first StartEvent. - Range: 0 (no sync), 1 (sync)
			1,	// Resampler or DDC			Send a synchronization strobe to the resampler immediately before processing the first StartEvent. - Range: 0 (no sync), 1 (sync)
			0,	// Reserved					Reserved for future expansion.
			0	// Reserved					Reserved for future expansion.
		},
		{ // SyncCycleMask
			0,	// Binary Data Counter		Send a synchronization strobe to the binary data counter immediately before processing all but the StartEvent of a cycle. - Range: 0 (no sync), 1 (sync)
			0,	// Resampler				Send a synchronization strobe to the resampler immediately before processing all but the StartEvent of a cycle. - Range: 0 (no sync), 1 (sync)
			0,	// Reserved					Reserved for future expansion.
			0	// Reserved					Reserved for future expansion.
		}
	},
	{ // Synchronizer
		1,	// StartEvent					Event that will activate the channel. - Range: 0 (none/immediate), 1 (software trigger), 2 (Coax trigger rising), 3 (Coax trigger falling), 4 (reserved), 5 (TOD), 6 (GPIO trigger rising), 7 (GPIO trigger falling)
		0,	// StartOffset					Number of sample clock periods to delay activating the channel after the start event is detected. - Range: 0 to 2^16 - 1
		4,	// StopEvent					Event that will deactivate the channel. - Range: 0 (flow control), 1 (software trigger), 2 (Coax trigger rising), 3 (Coax trigger falling), 4 (sample count), 5 (TOD), 6 (GPIO trigger rising), 7 (GPIO trigger falling)
		0,	// StopOffset					Number of sample clock periods to delay the actual deactivation after the stop event is detected. - Range: 0 to 2^16 - 1
		0,	// CycleSelect					Event that will initiate the start of a new cycle. - Range: 0 (none/single cycle), 1 (start event), 2 (frame), 3 (scheduler)
		0	// CycleStop					Event that will discontinue cycling. - Range: 0 (flow control), 1 (cyle count), 2 (TOD)
	},
	{ // TODTriggers
		0,	// IntegerSecondsStart			Integer seconds value to use when TOD is selected for the synchronizer StartEvent. - Range: 0 to 2^32 - 1
		0,	// FractionalSecondsStart		Fractional seconds value to use when TOD is selected for the synchronizer StartEvent. - Range: 0 to 2^32 - 1
		0,	// IntegerSecondsStop			Integer seconds value to use when TOD is selected for the synchronizer StopEvent or CycleStop. - Range: 0 to 2^32 - 1
		0	// FractionalSecondsStop		Fractional seconds value to use when TOD is selected for the synchronizer StopEvent or CycleStop. - Range: 0 to 2^32 - 1
	},
	{ // EventDuration;
		0x00004000,	// SampleCount			Number of samples to collect when sample count is selected for the synchronizer StartEvent. - Range: 0 to (2^32 - 1) in multiples of 8 bytes if data is not packetized
		0,	// Reserved						Reserved for future expansion.
		1,	// CycleCount					Number of cycles to execute when cycle count is selected for the synchronizer CycleStop. - Range: 1 to 2^32 - 1
		0	// FrameCount					Number of sample clock periods that define a frame duration when frame is selected for the synchronizer CycleSelect. - Range: 4 to 2^32 - 1
	},
	{ // IFdpktFormat
		0,	// PacketSize					Total number of 32-bit words present in the IF data packet as defined by VITA 49. - Range: 0 (data flow without packets) to 2^16 - 1
		0,	// TSFSelect					Insert a fractional seconds timestamp in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// TSISelect					Insert an integer seconds timestamp in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// TrailerSelect				Trailer appended to the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// ClassIDSelect				Insert a Class Identifier in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0ull,	// ClassID					Value of the optional IF data packet Class Identifier as defined by VITA 49. - Range: 0 to 2^64 - 1 (Note: Must include "ull" suffix on value.)
		0,	// StreamIDSelect				Insert a Stream Identifier in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// StreamID						Value of the optional IF data packet Stream Identifier as defined by VITA 49. - Range: 0 to 2^32 - 1
		0,	// ContextPktCntEnable			Enable the "Associated Context Packet Count" field of the IF packet trailer as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// IndicatorEnables				Enable bits associated with the twelve state and event indicators in the IF packet trailer as defined by VITA 49. - Range: 0 to 2^12 - 1
		0,	// SpecralInversion				Set the "Spectral Inversion" state and event indicator bit for the IF packet trailer as defined by VITA 49. - Range: 0 (non-inverted spectrum) to 1 (inverted spectrum)
		{0, 0, 0}	// UserIndicators		Three bits of user defined state and event indicators to include in the IF packet trailer as defined by VITA 49. - Range: 0 or 1 for each bit
	},
};

////////////////////////////////////////////  Configure Transmitter (TX) Channels  /////////////////////////////////////////////

// Initialize a default structure for the TX channels.
s_Channel TXDefault = {
	{ // SwitchSelect
		0,	// ChannelSelect				Select the Signal Stream (SS) channel for this transmitter (TX) output. - Range: 0 (product default), 1 (SS#1), 2 (SS#2), 3 (SS#3), 4 (SS#4), 5 (constant), 6 (training), 7 (RX Loopback)
											// Product default value: TX#1 = SS#1, TX#2 = SS#2, TX#3 = SS#3, TX#4 = SS#4
		0	// Amplitude					Amplitude of constant data source. - Range: 0 to 2^32 - 1
	},
	{ // PayloadFormat
		0,	// PayloadPacking				Data is packed in the DMA buffer as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// DataItemSize					Size of a data item in bits as defined by VITA 49. - Range: 0 (product default) 2, 4, 6 ... (size of DAC output)
											// Product default value: Model 272 = 16
		0,	// EventTagOrder				Not used by TX channel.
		0,	// EventTagMask					Not used by TX channel.
		0,	// EventTagIndex				Not used by TX channel.
		0,	// ItemPackingFieldSize			Size of the item packing field as defined by VITA 49 [ItemPackingFieldSize >= DataItemSize]. - Range: 4, 8, 10, 12, 16, 20, 24, 28, 32
											// Product default value: Model 272 = 16
		0	// ChannelTag					Not used by TX channel.
	},
	{ // DatapathControl
		{ // Transform
			0,	// Bandwidth Select			Optional filter/resampler to accept input bandwidths below Fs/2. - Range: 0 (BW=Fs/2 real), 1 (BW=Fs/2 complex), 2 (BW=Fs/4 complex), 3 (BW=Fs/8 complex) ... 11 (BW=Fs/2048 complex)
			0,	// Real Resampler			Replication factor (output : input) achieved by duplicating samples of real data only, not complex data. - Range: 0 (product default) to 2^16 - 1
			0,	// Frequency Translation	
			0,	// Phase Offset				
			0,	// Reserved
			0	// Reserved
		},
		{ // Residue
			0,	// Channel Stop				Disposition of data residue trapped in the pipeline when the channel is deactivated. - Range: 0 to 7
			0,	// Channel Cycle			Disposition of data residue trapped in the pipeline between cycles. - Range: 0 to 7
		},
		{ // SyncInitMask
			0,	// Reserved					Reserved for future expansion.
			1,	// Resampler				Send a synchronization strobe to the resampler immediately before processing the first StartEvent. - Range: 0 (no sync), 1 (sync)
			0,	// Reserved					Reserved for future expansion.
			0	// Reserved					Reserved for future expansion.
		},
		{ // SyncCycleMask
			0,	// Binary Data Counter		Not used by TX channel.
			1,	// Resampler				Send a synchronization strobe to the resampler immediately before processing all but the StartEvent of a cycle. - Range: 0 (no sync), 1 (sync)
			0,	// Reserved					Reserved for future expansion.
			0	// Reserved					Reserved for future expansion.
		}
	},
	{ // Synchronizer
		1,	// StartEvent					Event that will activate the channel. - Range: 0 (none/immediate), 1 (software trigger), 2 (Coax trigger rising), 3 (Coax trigger falling), 4 (reserved), 5 (TOD), 6 (GPIO trigger rising), 7 (GPIO trigger falling), 8 (GPIO#1 rising), 9 (GPIO#2 rising), 10 (GPIO#3 rising), 11 (GPIO#4 rising), 12 (GPIO#1 falling), 13 (GPIO#2 falling), 14 (GPIO#3 falling), 15 (GPIO#4 falling)
		0,	// StartOffset					Number of sample clock periods to delay activating the channel after the start event is detected. - Range: 0 to 2^16 - 1
		4,	// StopEvent					Event that will deactivate the channel. - Range: 0 (flow control), 1 (software trigger), 2 (Coax trigger rising), 3 (Coax trigger falling), 4 (sample count), 5 (TOD), 6 (GPIO trigger rising), 7 (GPIO trigger falling), 8 (GPIO#1 rising), 9 (GPIO#2 rising), 10 (GPIO#3 rising), 11 (GPIO#4 rising), 12 (GPIO#1 falling), 13 (GPIO#2 falling), 14 (GPIO#3 falling), 15 (GPIO#4 falling)
		0,	// StopOffset					Number of sample clock periods to delay the actual deactivation after the stop event is detected. - Range: 0 to 2^16 - 1
		0,	// CycleSelect					Event that will initiate the start of a new cycle. - Range: 0 (none/single cycle), 1 (start event), 2 (frame)
		0	// CycleStop					Event that will discontinue cycling. - Range: 0 (flow control), 1 (cyle count), 2 (TOD)
	},
	{ // TODTriggers
		0,	// IntegerSecondsStart			Integer seconds value to use when TOD is selected for the synchronizer StartEvent. - Range: 0 to 2^32 - 1
		0,	// FractionalSecondsStart		Fractional seconds value to use when TOD is selected for the synchronizer StartEvent. - Range: 0 to 2^32 - 1
		0,	// IntegerSecondsStop			Integer seconds value to use when TOD is selected for the synchronizer StopEvent. - Range: 0 to 2^32 - 1
		0	// FractionalSecondsStop		Fractional seconds value to use when TOD is selected for the synchronizer StopEvent. - Range: 0 to 2^32 - 1
	},
	{ // EventDuration;
		0x04000000,	// SampleCount			Number of samples to process when sample count is selected for the synchronizer StartEvent. - Range: 0 to (2^32 - 1) in multiples of 8 bytes if data is not packetized
		0,	// BurstCount					Number of DMA bursts sent from the FIFO to host memory. - Range: 0 to 2^32 - 1
		1,	// CycleCount					Number of cycles to execute when cycle count is selected for the synchronizer CycleStop. - Range: 1 to 2^32 - 1
		0	// FrameCount					Number of sample clock periods that define a frame duration when frame is selected for the synchronizer CycleSelect. - Range: 0 to 2^32 - 1
	},
	{ // IFdpktFormat
		0,	// PacketSize					Total number of 32-bit words present in the IF data packet as defined by VITA 49. - Range: 0 (data flow without packets) to 2^16 - 1
		0,	// TSF							Type of fractional seconds timestamp included in the data packet as defined by VITA 49.  - Range: 0 (none), 1 (sample count), 2 (real time), 3 (free running)
		0,	// TSI							Type of integer seconds timestamp included in the data packet as defined by VITA 49.  - Range: 0 (none), 1 (UTC), 2 (GPS), 3 (other)
		0,	// TrailerSelect				Trailer appended to the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// ClassIDSelect				Insert a Class Identifier in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0ull,	// ClassID					Not used by TX channel.
		0,	// StreamIDSelect				Insert a Stream Identifier in the IF data packet as defined by VITA 49. - Range: 0 (disable) to 1 (enable)
		0,	// StreamID						Not used by TX channel.
		0,	// ContextPktCntEnable			Not used by TX channel.
		0,	// IndicatorEnables				Not used by TX channel.
		0,	// SpecralInversion				Not used by TX channel.
		{0, 0, 0}	// UserIndicators		Not used by TX channel.
	},
};

///////////////////////////////////////////////////  Configure Interrupts  ////////////////////////////////////////////////////

// Initialize a default variable for the global interrupt mask setting.
unsigned long GlobalMaskDefault = 0;	// Global Interrupt Mask 	Enable or disable all interrupt generation from the device. - Range: 0 (disable) to 1 (enable)

// Initialize a default structure for the fault interrupt mask.
// Note: The number of active SRAM or ADC error indicators depends on specific product features.
// Note: The Alert indicator is only relevant if the port expander (FXL6408) or temperature monitor (LTC2991) are programmed to report status.
// Note: The Fan error indicator is only relevant on products equipped with an active heat sink.
s_ErrorStatus ErrorMaskDefault = {
	{ // BAR Error
		0,	// BAR0 Write Error			Generate an interrupt when a BAR0 write error is detected. - Range: 0 (disable) to 1 (enable)
		0,	// BAR0 Read Error			Generate an interrupt when a BAR0 read error is detected. - Range: 0 (disable) to 1 (enable)
		0,	// BAR2 Write Error			Generate an interrupt when a BAR2 write error is detected. - Range: 0 (disable) to 1 (enable)
		0	// BAR2 Read Error			Generate an interrupt when a BAR2 read error is detected. - Range: 0 (disable) to 1 (enable)
	},
	{ // Clock
		0,	// IDELAY Lock				Generate an interrupt when the FPGA IDELAY controller is out of lock. - Range: 0 (disable) to 1 (enable)
		0,	// Flash DCM Lock			Generate an interrupt when the Flash DCM is out of lock. - Range: 0 (disable) to 1 (enable)
		0,	// SRAM-A Lock				Generate an interrupt when the SRAM-A clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
		0,	// SRAM-B Lock				Generate an interrupt when the SRAM-B clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
		0,	// ADC Lock					Generate an interrupt when the ADC clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
		0,	// ADC Lock					Generate an interrupt when the ADC clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
		0,	// ADC Lock					Generate an interrupt when the ADC clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
		0	// ADC Lock					Generate an interrupt when the ADC clock lock function produces an error. - Range: 0 (disable) to 1 (enable)
	},
	0,		//Alert						Generate an interrupt when the port expander or temperature monitor alert is active. - Range: 0 (disable) to 1 (enable)
	0		//Fan						Generate an interrupt when the fan monitor reports an error. - Range: 0 (disable) to 1 (enable)
};

// Initialize a default structure for the channel status interrupt mask.
s_ChannelIntMask ChannelMaskDefault = {
	1,	// DMA Marker						Generate an interrupt when the DMA marker is detected. - Range: 0 (disable) to 1 (enable)
	{ // Error
		0,	// Converter (ADC/DAC) Error	Generate an interrupt when a converter (ADC/DAC) error is detected. - Range: 0 (disable) to 1 (enable)
		0	// FIFO Error					Generate an interrupt when a FIFO error is detected. - Range: 0 (disable) to 1 (enable)
	}
};

#ifdef __cplusplus
}
#endif
