/******************************************************  DMA Functions  *******************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to control and monitor the DMA engines.

	Rev History:  03-07-2019:	Corrected stripping of DMA engine status bits in DMAStatus().
	Rev History:  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************  DMA Functions  *******************************************************/

#include <math.h>
#include <stdio.h>
#include <errno.h>

#include "driver_functions.h"
#include "sigstream.h"
#include "global_functions.h"
#include "channel_functions.h"
#include "dma_functions.h"

// Assign DMA descriptors and physical page addresses to a device channel.
void DMAAssign(s_Handle *p_HW, unsigned BufferIndex)
{
	unsigned pagecnt;
	unsigned long long descriptor;
	unsigned addindex;
	
	// Set the DMA descriptors register index to load page offsets for the selected channel.
	p_HW->DEV.BAR0[DMA_DESCRIPTORS - 1] = 0x2 << BufferIndex;
	
	// Initialize the address index to the start of the page memory.
	addindex = DMA_DESCRIPTORS;
			
	// Write the DMA physical page addresses to the adapter.
	for(pagecnt=0; pagecnt < p_HW->DMA[BufferIndex].PageCount; pagecnt++)
	{
		// Perform the write operation to the page memory.
		p_HW->DEV.BAR0[addindex] = p_HW->DMA[BufferIndex].Address[pagecnt];
		// Increment the page memory address.
		addindex += 1;
	}
	
	// Set the DMA descriptors register index to load descriptors for the selected channel.
	p_HW->DEV.BAR0[DMA_DESCRIPTORS - 1] = 0x200 << BufferIndex;
			
	// Initialize the address index to the start of the descriptor memory.
	addindex = DMA_DESCRIPTORS;
	
	// Form the descriptors and write them to the adapter.
	for(pagecnt=0; pagecnt < p_HW->DMA[BufferIndex].PageCount; pagecnt++)
	{
		// Set the Burst Size and Burst Count fields of the descriptor.
		descriptor = (unsigned long long)(p_HW->DMA[BufferIndex].BurstSizeB << 16) | (p_HW->DMA[BufferIndex].BurstCount);
	
		//Set the Last Page descriptor bit if this is the final page of this channel.
		if (pagecnt == p_HW->DMA[BufferIndex].PageCount - 1)
			descriptor |= 0x0000000040000000ull;
	
		//Set the Interrupt descriptor bit if this page matches the requested pages per interrupt setting.
		if (((pagecnt+1) % p_HW->DMA[BufferIndex].PagesPerMark) == 0)
			descriptor |= 0x0000000080000000ull;
		// Perform the write operation to the descriptor memory.
		p_HW->DEV.BAR0[addindex] = descriptor;
		// Increment the descriptor memory address.
		addindex += 1;
	}
}

// Verify that the DMA configuration settings have legal values.
// Return of non-zero indicates a configuration error.
int DMAConfigCheck(s_DMABuf DMA, unsigned *ConfigError)
{
	unsigned pagecnt;
	unsigned index = 0;
	
	for(pagecnt=0; pagecnt < DMA.PageCount; pagecnt++)
	{
		// Determine whether the burst size is within range and an even multiple of the byte alignment.
		if ((DMA.BurstSizeB > MAX_BURSTSIZEB) || fmod(DMA.BurstSizeB, BYTE_ALIGNMENT)) ConfigError[0]++;

		// Determine whether the burst count is within range.
		if (DMA.BurstCount > MAX_BURSTCOUNT) ConfigError[1]++;
	}

	// Determine whether the page count per channel is within range.
	if (DMA.PageCount > MAX_PAGECOUNT) ConfigError[2]++;

	// Determine whether the PagesPerMark divides evenly into the PageCount.
	if (fmod(DMA.PageCount, DMA.PagesPerMark)) ConfigError[3]++;
		
	for (index = 0; index < 4; index++)
		if (ConfigError[index]) return DMA_BUFFER;

	return OK;
}

// Preload a DMA buffer with data samples from a file.
// Return non-zero indicates a failure opening the sample data text file to read.
int DMALoad(char *p_File, s_DMABuf *DMA, unsigned DataItemSize)
{
	FILE *fp;
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	unsigned page_count = 0;
	unsigned record_count = 0;
	unsigned data_items_count = 0;
	int sample = 0;

	// Reset the DMA page counter so the next mapping will call the first page in the buffer. 
	DMAFirstPage(DMA);
	
	// Sequence through all of the number of pages requested.
	for(page_count = 0; page_count < DMA->PageCount; page_count++)
	{
		// Open the requested file and read enough samples to fill a single DMA page.
		if((fp = fopen(p_File, "r")) == NULL)
			return errno;
		else
		{
			// Map the virtual address of the next DMA page.
			if (DMAMap(DMA))
				return errno;
				
			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < (DMA->BurstSizeB * DMA->BurstCount) / 8; record_count++)
			{
				// Clear the current 64-bit record.
				DMA->Address[record_count] = 0x0ull;
				
				for(data_items_count = 0; data_items_count < 64 / DataItemSize; data_items_count++)
				{
					// Read the next sample from the file.
					if (fscanf(fp, "%d", &sample) != 1)
						return EIO;
					
					// Write the sample as the current data item to the payload record.
					DMA->Address[record_count] |= ((max >> (64-DataItemSize)) & (unsigned long long)sample) << data_items_count*DataItemSize;
				}
			}
		// Unmap the current DMA page.
		if (DMAUnmap(DMA))
			return errno;
		// Close the file.
		fclose(fp);
		}
	}
	return OK;
}

// Write the contents of the DMA buffer to the specified file from the first page to the number of pages requested.
// Return of non-zero indicates indicates a failure opening the output data text file to write.
int DMASave(char *p_File, s_DMABuf *DMA, s_Channel RXChannel, s_TODSettings TODClock, unsigned NumPages, unsigned PacketInfo)
{
	FILE *fp;
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	unsigned current_field = 0;
	unsigned long field;
	unsigned field_count = 0;
	unsigned index = 0;
	unsigned next_field[8] = {0,0,0,0,0,0,0,0};
	unsigned packet_counter = 0;
	unsigned page_count = 0;
	unsigned payload_index = 32;
	unsigned record_count = 0;
	unsigned long residue = 0;
	int sample = 0;

	// Data will be packetized if Packet Size is not zero.
	if (RXChannel.IFdpktFormat.PacketSize)
	{
		// Create a linked list of expected packet record types based on the packet header options selected.
		if (RXChannel.IFdpktFormat.StreamIDSelect) {next_field[index] = 1; index = 1;}
		if (RXChannel.IFdpktFormat.ClassIDSelect) {next_field[index] = 2; index = 3;}
		if (RXChannel.IFdpktFormat.TSISelect) {next_field[index] = 4; index = 4;}
		if (RXChannel.IFdpktFormat.TSFSelect) {next_field[index] = 5; index = 6;}
		next_field[index] = 7;
	}
	// Data is not packetized, begin processing samples immediately.
	else
		current_field = 7;
	
	// Process records from the buffer only if the file open was successful.
	if((fp = fopen(p_File, "w")) == NULL)
		return errno;
	else
	{
		// Sequence through the pages.
		for(page_count=0; page_count < NumPages; page_count++)
		{
			// Map the virtual address of the next DMA page.
			if (DMAMap(DMA))
				return errno;
			
			// Sequence through every 64-bit record in the page.
			for(record_count = 0; record_count < (DMA->BurstSizeB * DMA->BurstCount) / 8; record_count++)
				// There are two 32-bit VITA 49 fields per 64-bit record.
				for(field_count = 0; field_count < 2; field_count++)
				{
					// Select the upper or lower 32 bits of the 64-bit record for the current field.
					if (field_count)
						// Bits 63:32 contain the next field.
						field = (unsigned long)(DMA->Address[record_count] >> 32);
					else
						// Bits 31:0 contain the next field.
						field = (unsigned long)DMA->Address[record_count];

					// Process the current record based on the type of information it conveys.
					switch (current_field)
					{
						// Process the Header field of a packet. (The current record must be a valid header or it will be discarded until a valid header is found.)
						case 0:
						{
							// Note: This bit pattern is not unique to a header, it is possible (but unlikely) for a data sequence to produce an unexpected match.
							if	(((field & 0xFFFF) == RXChannel.IFdpktFormat.PacketSize) &&		// Check for a match to the Packet Size field.
							((RXChannel.IFdpktFormat.TSFSelect == 1 && ((field >> 20) & 0x3) == TODClock.TSFCode) ||	// Check for a match to the TSF field.
								(RXChannel.IFdpktFormat.TSFSelect == 0 && ((field >> 20) & 0x3) == 0)) &&				// Check for a TSF field set to zero.
							((RXChannel.IFdpktFormat.TSISelect == 1 && ((field >> 22) & 0x3) == TODClock.TSICode) ||	// Check for a match to the TSI field.
								(RXChannel.IFdpktFormat.TSISelect == 0 && ((field >> 22) & 0x3) == 0)) &&				// Check for a TSI field set to zero.
							(((field >> 24) & 0x3) == 0) &&										// Check for zeros in the reserved field.
							(((field >> 26) & 0x1) == RXChannel.IFdpktFormat.TrailerSelect) &&	// Check for a match to the Trailer field.
							(((field >> 27) & 0x1) == RXChannel.IFdpktFormat.ClassIDSelect) &&	// Check for a match to the Class ID field.
							(((field >> 28) & 0x1) == RXChannel.IFdpktFormat.StreamIDSelect) &&	// Check for a match to the Stream ID field.
							(((field >> 29) & 0x0) == 0))										// Check for a match to the Packet Type field.
							{
								// Print header information to the file.
								if (PacketInfo) fprintf(fp, "Packet Number:  %lu\n", (field >> 16) & 0xF);
								packet_counter = 1;
								current_field = next_field[0];
								residue = 0;
								payload_index = 32;
							}
							break;
						}

						// Process the optional Stream Identifier field of a packet.
						case 1: {if (PacketInfo) fprintf(fp, "Stream Identifier:  x%08lX\n", field); packet_counter++; current_field = next_field[1]; break;}

						// Process the first half of the optional Class Identifier field of a packet.
						case 2: {if (PacketInfo) fprintf(fp, "Class Identifier:  x%08lX", field); packet_counter++; current_field = 3; break;}

						// Process the second half of the optional Class Identifier field of a packet.
						case 3: {if (PacketInfo) fprintf(fp, "%08lX\n", field); packet_counter++; current_field = next_field[3]; break;}

						// Process the optional Integer-seconds Timestamp field of a packet.
						case 4: {if (PacketInfo) fprintf(fp, "Integer-seconds Timestamp:  x%08lX\n", field); packet_counter++; current_field = next_field[4]; break;}

						// Process the first half of the optional Fractional-seconds Timestamp field of a packet.
						case 5: {if (PacketInfo) fprintf(fp, "Fractional-seconds Timestamp:  x%08lX", field); packet_counter++; current_field = 6; break;}

						// Process the second half of the optional Fractional-seconds Timestamp field of a packet.
						case 6: {if (PacketInfo) fprintf(fp, "%08lX\n", field); packet_counter++; current_field = next_field[6]; break;}

						// Process the Data Payload field of a packet or a raw stream.
						case 7:
						{
							// Sequence through every sample in the 32-bit payload.
							while (payload_index >= RXChannel.PayloadFormat.ItemPackingFieldSize)
							{
								// Decrement the payload index to extract the next sample in the 32-bit payload.
								payload_index -= (unsigned)RXChannel.PayloadFormat.ItemPackingFieldSize;
								// Extract the next sample value from the 32-bit payload and append any packing residue that may be present from the last payload.
								sample = (int)(((field >> payload_index) & (0xFFFFFFFF >> (32 - RXChannel.PayloadFormat.ItemPackingFieldSize))) | residue);
								// Clear the residue for the remainder of the current payload.
								residue = 0;
								// Sign extend negative numbers.
								if (sample & (one << (RXChannel.PayloadFormat.ItemPackingFieldSize - 1))) sample |= (max << (RXChannel.PayloadFormat.ItemPackingFieldSize - 1));
								// Print the sample value to the file.
								fprintf(fp, "%d\n", sample);
							}

							// Extract the residue and adjust the payload index if packing is enabled and there is residue present in the current payload.
							if (RXChannel.PayloadFormat.PayloadPacking && payload_index)
							{
								residue = (field & (0xFFFFFFFF >> (32 - payload_index))) << (RXChannel.PayloadFormat.ItemPackingFieldSize - payload_index);
								payload_index += 32;
							}
							// Clear the residue and reset the payload index if packing is not enabled or there is no residue present in the current payload.
							else
							{
								residue = 0;
								payload_index = 32;
							}

							// Track the number of payloads processed if packets are enabled (Packet Size != 0).
							if (RXChannel.IFdpktFormat.PacketSize)
							{
								packet_counter++;
								// Determine whether this is the last payload in a packet when the optional trailer is selected.
								if (RXChannel.IFdpktFormat.TrailerSelect && (packet_counter == (RXChannel.IFdpktFormat.PacketSize - 1)))
									current_field = 8;
								// Determine whether this is the last payload in a packet when the optional trailer is not selected.
								else if (packet_counter == RXChannel.IFdpktFormat.PacketSize)
									current_field = 0;
							}
							break;
						}

						// Process the optional Trailer field of a packet.
						case 8: {if (PacketInfo) fprintf(fp, "Trailer:  x%08lX\n", field); packet_counter++; current_field = 0; break;}
					}
				}
				
			// Unmap the current DMA page.
			if (DMAUnmap(DMA))
				return errno;
		}
		// Close the file.
		fclose(fp);
		return OK;
	}
}

// Report the current DMA on Demand operating status.
void DMAStatus(s_Handle *p_HW, unsigned ChannelNumber, s_DMAStatus *p_DMA)
{
	unsigned long long value = 0;
	unsigned cnt = 0;
	
	// Read the requested DMA status address.
	switch (ChannelNumber)
	{
		case 1: {value = p_HW->DEV.BAR0[DMA_1TO2_STATUS]; break;}
		case 2: {value = p_HW->DEV.BAR0[DMA_1TO2_STATUS]; value >>= 32; break;}
		case 3: {value = p_HW->DEV.BAR0[DMA_3TO4_STATUS]; break;}
		case 4: {value = p_HW->DEV.BAR0[DMA_3TO4_STATUS]; value >>= 32; break;}
		case 5: {value = p_HW->DEV.BAR0[DMA_5TO6_STATUS]; break;}
		case 6: {value = p_HW->DEV.BAR0[DMA_5TO6_STATUS]; value >>= 32; break;}
		case 7: {value = p_HW->DEV.BAR0[DMA_7TO8_STATUS]; break;}
		case 8: {value = p_HW->DEV.BAR0[DMA_7TO8_STATUS]; value >>= 32; break;}
		default: break;
	}

	// Strip the current page.
	p_DMA->CurrentPage = value & 0x00000000000003FFull;

	// Strip the current burst.
	p_DMA->CurrentBurst = (value >> 16) & 0x000000000000FFFFull;

	// Strip the individual DMA engine status bits (0 = Enable, 1 = Go, 2 = Busy).
	for (cnt = 0; cnt < 3; cnt++)
		p_DMA->Engine[cnt] = (value >> (cnt + 10)) & 0x0000000000000001ull;
}
