/****************************************************  Filter Functions  *****************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to configure a variety of filters available on SigStream products.

	Rev History:  01-20-2020:	Corrected bit shift on all filter configuration command operations.
				  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Filter Functions  *****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include "driver_functions.h"
#include "sigstream.h"
#include "global_functions.h"
#include "channel_functions.h"
#include "dma_functions.h"
#include "filter_functions.h"

// Define the unique characteristics of every filter type available in the product family.
unsigned Filter[FilterTypes][6] =
{
	// Type 0 FIR filter definition.
	{
		87,		// Number of Taps			Size of the filter in number of taps or coefficients.
		0,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		24,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
	},
	// Type 1 FIR filter definition.
	{
		127,	// Number of Taps			Size of the filter in number of taps or coefficients.
		1,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		24,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
	},
	// Type 2 FIR filter definition.
{
		255,	// Number of Taps			Size of the filter in number of taps or coefficients.
		2,		// Reload Order				Unique code to convey the coefficient reload order of the filter.
		23,		// Coeffiecient Size		Size of each coefficient in bits.
		1,		// Symmetry					Code to convey symmetry properties of the filer (0 = asymmetric, 1 = symmetric, 2 = antisymmetric).
		1,		// Minimum Resample			Minimum downsample ratio (input:output) or upsample ratio (output:input).
		0		// Reserved
}
};

// Load the filter coefficients.
// Return of non-zero indicates an unknown filter type or command error.
int FilterCoefficients(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select, unsigned Type, unsigned *p_Coefficients)
{
	unsigned long long value = 0;
	unsigned index = 0;
	unsigned symmetry = 2;

	// Determine if coefficients are asymmetric (0).
	for (index = 0; index < Filter[Type][0]/2; index++)
		if ((p_Coefficients[index] != p_Coefficients[Filter[Type][0]-1-index]) && (p_Coefficients[index] != -1* p_Coefficients[Filter[Type][0]-1-index]))
			symmetry = 0;
	// If the coefficients are not asymmetric, determine wheter they are symmetric (1) or antisymmetric (2).
	if ((symmetry != 0) && (p_Coefficients[0] == p_Coefficients[Filter[Type][0]-1]))
		symmetry = 1;

	// Make sure the supplied coefficients match the characteristics.
	if (symmetry == Filter[Type][3])
		// Reorder the odd coefficents and reload the FIR with symmetry/anitsymmetry.
		if (symmetry)
			for (index = 0; index <= Filter[Type][0]/2; index++)
			{
				// Read the filter configuraton status address to determine whether there is a command pending, which should be impossible.
				value = p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_STAT];
				// Load the next coefficient if there is no command pending.
				if (!(value & 0x0000000000000100ull))
				{
					// Select the next coefficient based on a specific reload order.
					switch (Filter[Type][1])
					{
						// Forward sequential.
						case 0: { value = p_Coefficients[index]; break; }
						// Reverse modulo 2.
						case 1: { value = p_Coefficients[((Filter[Type][0] + 1) / 2) - 2 - index + (2 * (index % 2))]; break; }
						// Reverse modulo 4.
						case 2: { value = p_Coefficients[((Filter[Type][0] + 1) / 2) - 4 - index + (2 * (index % 4))]; break; }
						// Abort if there is no match.
						default: { return COEFFICIENT_ERROR; }
					}
					// Write the coefficient reload value.
					if (index + 1 == Filter[Type][0]/2)
						// Write to the filter configuration data address with the tlast bit enabled.
						p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = 0x0000000001000000ull | value;
					else
						// Write to the filter configuration data address with the tlast bit disabled.
						p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = value;
					// Write the command to execute the reload operation of the FIR to the filter configuration command address.
					p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000001ull;
				}
				// There is a command pending which shouldn't be possible so return an error.
				else
					return COEFFICIENT_ERROR;
			}
		// Reorder the even coefficents and reload the FIR with no symmetry.
		else
			for (index = 0; index < Filter[Type][0]; index++)
			{
				// Read the filter configuraton status address to determine whether there is a command pending, which should be impossible.
				value = p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_STAT];
				// Load the next coefficient if there is no command pending.
				if (!(value & 0x0000000000000100ull))
				{
					// Select the next coefficient based on a specific reload order.
					switch (Filter[Type][1])
					{
						// Forward modulo 32.
						case 4: { value = p_Coefficients[Filter[Type][0] - 32 + (index % 32) - (index / 32) * 32]; break; }
						// Abort if there is no match.
						default: { return COEFFICIENT_ERROR; }
					}
					// Write the coefficient reload value.
					if (index == Filter[Type][0]-1)
						// Write to the filter configuration data address with the tlast bit enabled.
						p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = 0x0000000001000000ull | value;
					else
						// Write to the filter configuration data address with the tlast bit disabled.
						p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = value;
					// Write the command to execute the reload operation of the FIR to the filter configuration command address.
					p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000001ull;
				}
				// There is a command pending which shouldn't be possible so return an error.
				else
					return COEFFICIENT_ERROR;
			}
	// Return an error if the coefficents don't match the expected symmetry.
	else
		return COEFFICIENT_ERROR;

	// Write a zero config value to the filter configuration data adderess since the filter only supports one coefficent set.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = 0x0000000000000000ull;
	// Activate the latest coefficient reload values by writing to the filter configuration command address.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000003ull;

	return OK;
}

// Load coefficients from a file into the designated filter.
// Return of non-zero indicates a memory allocation error.
int FilterConfigure(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select, char *p_File)
{
	int error_code = 0;
	unsigned *p_coefficients;
	unsigned log2gain = 0;
	unsigned resample = 0;
	unsigned type = 0;

	// Get the type of the selected filter.
	type = FilterType(p_HW, ChannelNumber, Select);

	// Write to the filter configuration command address to select the FIR but not execute a command (LSB = 0).
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000000ull;

	// Allocate storage for all of the coefficients.
	if ((p_coefficients = malloc(sizeof *p_coefficients * Filter[type][0])) == NULL)
		return errno;

	// Parse the configuration file for filter characteristics.
	if ((error_code = FilterParser(type, p_coefficients, &log2gain, &resample, p_File)))
		return error_code;

	// Load the filter coefficients.
	if ((error_code = FilterCoefficients(p_HW, ChannelNumber, Select, type, p_coefficients)))
		return error_code;

	// Load the filter gain setting.
	if ((error_code = FilterGain(p_HW, ChannelNumber, Select, type, log2gain)))
		return error_code;

	// Load the filter resampler setting.
	if ((error_code = FilterResampler(p_HW, ChannelNumber, Select, type, resample)))
		return error_code;

	// Free the coefficient memory.
	free(p_coefficients);

	return OK;
}

// Report the number of filters assigned to the requested channel.
// Return the 8-bit filter count.
unsigned FilterCount(s_Handle *p_HW, unsigned ChannelNumber)
{
	// Read the filter configuration status address.
	return (unsigned)((p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_STAT] >> 32) & 0x00000000000000FFull);
}

// Create a coefficient file name of the form filter<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
void FilterFilename(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select, char *p_File)
{
	unsigned type = 0;
	char number[3];

	// Get the type of the selected FIR.
	type = FilterType(p_HW, ChannelNumber, Select);

	// Create a coefficient ch0 of the form filter<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
	// Note: Assumes files are located in directory ..\coefficients (Windows) or ../coefficents (Linux).
	strcpy(p_File, "..");
	strcat(p_File, SEPARATOR);
	strcat(p_File, "filters");
	strcat(p_File, SEPARATOR);
	strcat(p_File, "filter");
	sprintf(number, "%u", Select);
	strcat(p_File, number);
	strcat(p_File, "_type");
	sprintf(number, "%u", type);
	strcat(p_File, number);
	strcat(p_File, ".txt");
}

// Load the filter gain.
// Return of non-zero indicates requested gain is out of range.
int FilterGain(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select, unsigned Type, unsigned Log2Gain)
{
	unsigned long long bit_shift = 0;

	// Set the bit shift equal to the maximum number of coefficent magnitude bits, not including the sign bit.
	bit_shift = Filter[Type][2] - 1;

	// Return an error if the bit shifter is not within the range of zero to seven bits from full scale.
	if ((Log2Gain < Filter[Type][2] - 1) || (Log2Gain > Filter[Type][2] + 8))
		return GAIN_ERROR;
	else
		// Select the significant bits from the 24-bit output that will result in maximum dynamic range.
		bit_shift = Log2Gain - bit_shift;

	// Write to the filter configuration data address to set the bit shifter value to output the most significant MSBs.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = bit_shift;

	// Write to the filter configuration command address to load the gain value.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000005ull;

	return OK;
}

// Search for a keyword in the configuration file.
// Return the identification number assigned to the keyword.
unsigned FilterKeyword(char *Keyword)
{
	unsigned keyword_id = 0;

	// Create a coefficient ch0 of the form fir<#1>_type<#2>.txt, where #1 is the number and #2 is the type.
	if (!strncmp(Keyword, "Coef", 4) || !strncmp(Keyword, "coef", 4) || !strncmp(Keyword, "COEF", 4)) keyword_id = 1;
	if (!strncmp(Keyword, "Log2", 4) || !strncmp(Keyword, "log2", 4) || !strncmp(Keyword, "LOG2", 4)) keyword_id = 2;
	if (!strncmp(Keyword, "Resa", 4) || !strncmp(Keyword, "resa", 4) || !strncmp(Keyword, "RESA", 4)) keyword_id = 3;
	
	return keyword_id;
}

// Search for a keyword in the configuration file.
// Return of non-zero indicates file access error.
int FilterParser(unsigned Type, unsigned *p_Coefficients, unsigned *p_Log2Gain, unsigned *p_Resample, char *p_File)
{
	FILE *fp;
	unsigned cnt = 0;
	char inputline[16];

	// Process the coefficient file only if it opens successfully.
	if((fp = fopen(p_File, "r")) == NULL)
		return errno;
	else
	{
		// Parse the configuration file for keywords.
		while (fgets(inputline, 16, fp) != NULL)
			// Process the keywords.
			switch (FilterKeyword(inputline))
			{

				// Read the coefficients.
				case 1:
				{
					// Read and store all of the available coefficients.
					for (cnt = 0; cnt < Filter[Type][0]; cnt++)
						if (fscanf(fp, "%u", &p_Coefficients[cnt]) == EOF)
							return PARSER_ERROR;
					break;
				}

				// Read the base 2 logarithm of the gain.
				case 2: {if (fscanf(fp, "%u", p_Log2Gain) == EOF) return PARSER_ERROR; break; }

				// Read the resample factor.
				case 3: {if (fscanf(fp, "%u", p_Resample) == EOF) return PARSER_ERROR; break; }

				// Do nothing if not a recognized keyword.
				default: {break; }
			}
		// Close the file.
		fclose(fp);
		return OK;
	}
}

// Load the filter resmapler.
// Return of non-zero indicates requested resampler value is out of range.
int FilterResampler(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select, unsigned Type, unsigned Resample)
{
	// Set the resampler to the requested value.
	if (Resample == 0)
		// Write to the filter configuration data address to bypass the entire filter stage.
		p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = 0x0000000000010000ull;
	// Return an error if the requested resampler value is not either the minimum required or a multiple of the minimum required.
	// Determine whether the requested resampler value is at least the minimum required.
	else if ((Resample < Filter[Type][4]) || (Resample%Filter[Type][4] != 0))
		return RESAMPLER_ERROR;
	else
		// Write to the filter configuration data address to set the resampler to a zero based value after removing the built-in minimum value.
		p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_DATA] = (Resample/Filter[Type][4])-1;

	// Write to the filter configuration data address to load the resampler value.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = (((unsigned long long)Select - 1) << 5) | 0x0000000000000007ull;

	return OK;
}

// Report the unique type ID associated with the selected filter.
// Return the 8-bit type ID.
unsigned FilterType(s_Handle *p_HW, unsigned ChannelNumber, unsigned Select)
{
	// Write to the filter configuration address to select the specific filter to identify.
	p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_CMD] = ((unsigned long long)Select - 1) << 5;
	// Read the filter configuration status address to get the filter ID.
	return (unsigned)((p_HW->DEV.BAR2[(512 * ChannelNumber) + FILTER_CONFIG_STAT] >> 10) & 0x00000000000000FFull);
}
