/****************************************************  Global Functions  ******************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions apply to all SigStream hardware and channel types.

	Rev History:  01-20-2020:	Added SigStreamID() to identify the product when multiple devices are present.
				  03-07-2019:	Added ErrorMask() and updated ErrorStatus() to include Alert and Fan bits.
				  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Global Functions  ******************************************************/

#include <time.h>
#include <errno.h>

#include "driver_functions.h"
#include "sigstream.h"
#include "chip_functions.h"
#include "global_functions.h"

// Report the number of receiver (0) or transmitter (1) channels available in the device.
// Returns the number of channels.
unsigned ChannelCount(s_Handle *p_HW, unsigned RxTxSelect)
{
	// Read the identification address.
	return (unsigned long)((p_HW->DEV.BAR2[IDENTIFICATION] >> (44 + 4*RxTxSelect)) & 0x000000000000000Full);
}

// Get the measured frequency of the transmitter clock.
// Returns the measured clock frequency.
unsigned ClockFrequency(s_Handle *p_HW)
{
	// Write to the clock frequency address to start the measurement.
	p_HW->DEV.BAR0[CLOCK_FREQUENCY] = 0x1ull;
	// Wait for one second to elapse.
	MicrosecondTimer(1000000);
	// Write to the clock frequency address to stop the measurement.
	p_HW->DEV.BAR0[CLOCK_FREQUENCY] = 0x0ull;
	
	// Return the measured clock frequency.
	return (unsigned)p_HW->DEV.BAR0[CLOCK_FREQUENCY];
}

// Synchronize the converter clocks.
// Return of zero indicates success, non-zero indicates an error condition.
int ClockInit(s_Handle *p_HW)
{
	unsigned long long value = 0;
	unsigned long long busy = 0;
	unsigned long delay = 0;
	unsigned watchdog = 0;

	// Wait for clock lock to find a solution across the available search window.
	do
	{
		// Reset the watchdog counter.
		watchdog = 0;

		// Command the clock lock process to start by writing to the clock initialize address.
		p_HW->DEV.BAR0[CLOCK_INITIALIZE] = 0x0ull;

		// Wait for the clock lock process to begin.
		while (!busy && (watchdog < 100))
		{
			// Read the clock initialize address to check the clock lock busy status.
			busy = p_HW->DEV.BAR0[CLOCK_INITIALIZE];
			// Increment the watchdog counter.
			watchdog++;
		}

		// Clear the contents of the error status register.
		p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
			
		// Wait for clock lock to find a solution across the 2.5 ns search window.
		MicrosecondTimer(10);
		
		// Save and clear the contents of the error status register by writing to the error status address.
		// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
		p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
		// Read the saved copy of the error status to determine if all converter clocks are locked.
		value = p_HW->DEV.BAR0[ERROR_STATUS];

		// Phase shift the converter clock and start a new search if a solution was not found.
		// Note: Not all products support this feature.
		if ((ModelNumber(p_HW) != 0x376) && (value & 0x0000000000000F00ull))
		{
			// Disable the delay bypass in the clock distribution chip.
			AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x0034, 0x00);
			// Load the fine phase adjust with an integer from 0 to 31.
			// Note: Start from zero since the delay block includes an inherent 600 ps offset.
			AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x0036, delay << 1);
			// Perform a register update to commit the new phase delay values.
			AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x005A, 0x01);
			// Increment the delay approximately 300 ps for another possible search.
			delay++;
		}
	}
	while ((ModelNumber(p_HW) != 0x376) && (value & 0x0000000000000F00ull) && (delay != 32) && (watchdog < 100));

	// Return an error if no solution was found.
	if ((value & 0x0000000000000F00ull) || watchdog >= 100)
		return CLOCK_NOSOL;
	else
		return OK;
}

// Synchronize the converter clocks.
// Note: This version applies to products that use the LMK01000 clock distribution chip that does not support register read transactions.
// Return of zero indicates success, non-zero indicates an error condition.
int ClockInitLMK01000(s_Handle *p_HW, unsigned CLKout0, unsigned CLKout4, unsigned CLKout5)
{
	unsigned long long value = 0;
	unsigned long long busy = 0;
	unsigned long delay = 0;
	unsigned watchdog = 0;

	// Wait for clock lock to find a solution across the available search window.
	do
	{
		// Reset the watchdog counter.
		watchdog = 0;

		// Command the clock lock process to start by writing to the clock initialize address.
		p_HW->DEV.BAR0[CLOCK_INITIALIZE] = 0x0ull;

		// Wait for the clock lock process to begin.
		while (!busy && (watchdog < 100))
		{
			// Read the clock initialize address to check the clock lock busy status.
			busy = p_HW->DEV.BAR0[CLOCK_INITIALIZE];
			// Increment the watchdog counter.
			watchdog++;
		}

		// Clear the contents of the error status register.
		p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
			
		// Wait for clock lock to find a solution across the 2.5 ns search window.
		MicrosecondTimer(10);
		
		// Save and clear the contents of the error status register by writing to the error status address.
		// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
		p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
		// Read the saved copy of the error status to determine if all converter clocks are locked.
		value = p_HW->DEV.BAR0[ERROR_STATUS];

		// Apply a phase shift using the clock distribution chip and start a new search if a solution was not found.
		if (value & 0x0000000000000F00ull)
		{
			// Start by phase shifting the digital clock. 
			// Note: Start from zero since the delay block includes an inherent 400 ps offset.
			if (delay < 16)
				// Enable the delay on the digital clock (CLKout0) and load the delay with an integer from 0 to 15.
				LMK01000Write(p_HW, DEFAULT_CLK_DIST, 0x0, CLKout0 | 0x0004000 | delay);
			// Continue by phase shifting the ADC/DAC clocks. 
			else
			{
				// Remove the delay on the digital clock (CLKout0).
				LMK01000Write(p_HW, DEFAULT_CLK_DIST, 0x0, CLKout0);
				// Enable the delay on the DAC clock (CLKout4) and load the delay with an integer from 0 to 15.
				// Note: This clock is disabled by default on products that do not include a DAC.
				LMK01000Write(p_HW, DEFAULT_CLK_DIST, 0x4, CLKout4 | 0x0004000 | (delay-15));
				// Enable the delay on the ADC clock (CLKout5) and load the delay with an integer from 0 to 15.
				LMK01000Write(p_HW, DEFAULT_CLK_DIST, 0x5, CLKout5 | 0x0004000 | (delay-15));
			}
			// Increment the delay approximately 150 ps for the next search.
			delay++;
		}
	}
	while ((value & 0x0000000000000F00ull) && (delay != 32) && (watchdog < 100));

	// Return an error if no solution was found.
	if ((value & 0x0000000000000F00ull) || watchdog >= 100)
		return CLOCK_NOSOL;
	else
		return OK;
}

// Report the current status of all clocks.
// Returns the following status codes:
// 0: External sample clock active, digital clocks are locked.
// 1: External sample clock active, digital clocks are not locked.
// 2: Sample clock locked to internal reference, digital clocks are locked.
// 3: Sample clock locked to internal reference, digital clocks are not locked.
// 4: Sample clock locked to external reference, digital clocks are locked.
// 5: Sample clock locked to external reference, digital clocks are not locked.
// >5: Sample clock is not locked.
int ClockStatus(s_Handle *p_HW)
{
	unsigned long status = 0;
	
	if (!(FXL6408Read(p_HW, CMD_STAT1, 0xF) & 0x2))
		// External sample clock active.
		status = 0;
	else
		if (!(FXL6408Read(p_HW, CMD_STAT1, 0xF) & 0x1))
			//Sample clock not locked.
			status = 6;
		else
			if (FXL6408Read(p_HW, CMD_STAT1, 0xF) & 0x4)
				// Sample clock locked to internal reference.
				status = 2;
			else
				// Sample clock locked to external reference.
				status = 4;

	// Save and clear the contents of the error status register by writing to the error status address.
	// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
	p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
	// Read the saved copy of the error status and examine the IDELAY lock (4) and RX clock lock (11:8) error bits.
	if (p_HW->DEV.BAR0[ERROR_STATUS] & 0x0000000000000F10ull)
		// An error was reported on the digital clock status.
		status++;

	return status;
}

// Set the error interrupt mask bits to enable (1) or disable (0) interrupt generation for each error type.
void ErrorMask(s_Handle *p_HW, s_ErrorStatus ErrorIntMask)
{
	// Write to the error interrupt mask address.
	p_HW->DEV.BAR0[ERROR_INT_MASK] =
		(unsigned long long)ErrorIntMask.Bar[0] |
		(unsigned long long)ErrorIntMask.Bar[1] << 1 |
		(unsigned long long)ErrorIntMask.Bar[2] << 2 |
		(unsigned long long)ErrorIntMask.Bar[3] << 3 |
		(unsigned long long)ErrorIntMask.Clock[0] << 4 |
		(unsigned long long)ErrorIntMask.Clock[1] << 5 |
		(unsigned long long)ErrorIntMask.Clock[2] << 6 |
		(unsigned long long)ErrorIntMask.Clock[3] << 7 |
		(unsigned long long)ErrorIntMask.Clock[4] << 8 |
		(unsigned long long)ErrorIntMask.Clock[5] << 9 |
		(unsigned long long)ErrorIntMask.Clock[6] << 10 |
		(unsigned long long)ErrorIntMask.Clock[7] << 11 |
		(unsigned long long)ErrorIntMask.Alert << 12 |
		(unsigned long long)ErrorIntMask.Fan << 13;
}

// Report the error status and clear the values.
// Returns the total number of errors detected.
int ErrorStatus(s_Handle *p_HW, s_ErrorStatus *p_Error)
{
	unsigned long long value = 0;
	int status = 0;
	unsigned cnt = 0;
	
	// Save and clear the contents of the error status register by writing to the error status address.
	// Note: The contents are saved to a shadow register that is accessed during a read operation (below).
	p_HW->DEV.BAR0[ERROR_STATUS] = 0x0ull;
	// Read the saved copy of the error status.
	value = p_HW->DEV.BAR0[ERROR_STATUS];

	// Strip the individual BAR error indicator bits.
	for (cnt = 0; cnt < 4; cnt++)
		if ((p_Error->Bar[cnt] = (value >> cnt) & 0x0000000000000001ull)) status++;

	// Strip the individual clock error indicator bits.
	for (cnt = 0; cnt < 8; cnt++)
		if ((p_Error->Clock[cnt] = (value >> (cnt + 4)) & 0x0000000000000001ull)) status++;

	// Strip the Alert and Fan status bits.
	if ((p_Error->Alert = (value >> 12) & 0x0000000000000001ull)) status++;
	if ((p_Error->Fan = (value >> 13) & 0x0000000000000001ull)) status++;

	// Return the total error count.
	return status;
}

// Report the firmware revision date (MMDDYYYY).
// Returns the firmware revision date in hex format, but it is not a hex value (e.g. 04/27/2012 returns 0x04272012).
unsigned FirmwareRevision(s_Handle *p_HW)
{
	// Read the identification address.
	return (unsigned)(p_HW->DEV.BAR2[IDENTIFICATION] & 0xFFFFFFFFull);
}

// Report the state of channel and error status flags.
// Returns the total number of flags that require service.
int GlobalStatus(s_Handle *p_HW, s_GlobalStatus *p_Global)
{
	unsigned long long value = 0;
	unsigned long status = 0;
	unsigned cnt = 0;
	
	// Read the global status.
	value = p_HW->DEV.BAR0[GLOBAL_STATUS];

	// Query bit 7 to determine the bit mapping of the global status register.
	if (value & 0x0000000000000080ull)
	{
		// Strip the maximum number of possible individual channel status flags.
		for (cnt = 0; cnt < 8; cnt++)
			if ((p_Global->Channel[cnt] = (value >> (cnt + 16)) & 0x0000000000000001ull)) status++;
	
		// Strip the error status flag.
		if ((p_Global->Error = (value >> 15) & 0x0000000000000001ull)) status++;
	}
	else
	{
		// Strip the maximum number of possible individual channel status flags.
		for (cnt = 0; cnt < 4; cnt++)
			if ((p_Global->Channel[cnt] = (value >> (cnt * 4)) & 0x0000000000000001ull)) status++;
	
		// Strip the error status flag.
		if ((p_Global->Error = (value >> 32) & 0x0000000000000001ull)) status++;
	}

	// Return the total active flag count.
	return status;
}

// Issue a global synchronization strobe that is synchronous to all channels.
void GlobalSync(s_Handle *p_HW)
{
	// Write to the global synchronization address.
	p_HW->DEV.BAR2[GLOBAL_SYNC] = 0x0ull;
}

// Read GPIO input values and write GPIO output values as a six bit (5:0) binary format (GPIO #6 to GPIO #1).
// Return GPIO input values.
unsigned GPIOReadWrite(s_Handle *p_HW, unsigned GPIOWrite)
{
	// Write the GPIO output value address.
	p_HW->DEV.BAR2[GPIO_READ_WRITE] = (unsigned long long)GPIOWrite;

	// Read the GPIO input value address.
	return (unsigned)(p_HW->DEV.BAR2[GPIO_READ_WRITE] & 0x3F);
}

// Configure the bidirectional GPIO ports.
void GPIOSettings(s_Handle *p_HW, s_GPIOSettings GPIO)
{
	unsigned value = 0;
	unsigned direction = 0;

	// Tristate all of the voltage translators before changing the signal direction to avoid contention.
	p_HW->DEV.BAR2[GPIO_SETTINGS] = 0x1ull;

	// Set the GPIO voltage translator direction.
	direction =
		(GPIO.Direction[5] & 0x1) << 18 |
		(GPIO.Direction[4] & 0x1) << 19 |
		(GPIO.Direction[3] & 0x1) << 20 |
		(GPIO.Direction[2] & 0x1) << 21 |
		(GPIO.Direction[1] & 0x1) << 22 |
		(GPIO.Direction[0] & 0x1) << 23;

	// Load the translator direction settings into the port expander.
	Serial(p_HW, GPIO_USER_DIRECTION, RISING_EDGE, 0x0, 0, direction, 24, &value);

	// Load the GPIO configuration.
	p_HW->DEV.BAR2[GPIO_SETTINGS] =
		(unsigned long long)(GPIO.Disconnect & 0x1) |
		(unsigned long long)(GPIO.Direction[0] & 0x3) << 4 |
		(unsigned long long)(GPIO.Direction[1] & 0x3) << 6 |
		(unsigned long long)(GPIO.Direction[2] & 0x3) << 8 |
		(unsigned long long)(GPIO.Direction[3] & 0x3) << 10 |
		(unsigned long long)(GPIO.Direction[4] & 0x3) << 12 |
		(unsigned long long)(GPIO.Direction[5] & 0x3) << 14;
}

// Set the global interrupt mask to enable (1) or disable (0) interrupt generation from the device.
void InterruptMask(s_Handle *p_HW, unsigned GlobalIntMask)
{
	// Write to the global interrupt mask address.
	p_HW->DEV.BAR0[GLOBAL_INT_MASK] = (unsigned long long)GlobalIntMask;
}

// Select the sources that will illuminate LED-A and LED-B.
void LEDSettings(s_Handle *p_HW, s_LEDSettings LED)
{
	// Write to the LED source mask address.
	p_HW->DEV.BAR2[LED_SOURCE_MASK] =
		(unsigned long long)(LED.Yellow[0] & 0x1) << 1 |
		(unsigned long long)(LED.Yellow[1] & 0x3) << 2 |
		(unsigned long long)(LED.Yellow[2] & 0x3) << 4 |
		(unsigned long long)(LED.Yellow[3] & 0x3) << 6 |
		(unsigned long long)(LED.Green[0] & 0x1) << 16 |
		(unsigned long long)(LED.Green[1] & 0x1) << 17 |
		(unsigned long long)(LED.Green[2] & 0x1) << 18 |
		(unsigned long long)(LED.Green[3] & 0x1) << 19 |
		(unsigned long long)(LED.Green[4] & 0x1) << 20 |
		(unsigned long long)(LED.Green[5] & 0x1) << 21 |
		(unsigned long long)(LED.Green[6] & 0x1) << 22 |
		(unsigned long long)(LED.Green[7] & 0x1) << 23;
}

// Set a timer in units of microseconds.
// Return of zero indicates success, non-zero indicates an error condition.
int MicrosecondTimer(long uSeconds)
{
	struct timespec ts;
	
	// Convert the requested time to seconds and nanoseconds.
	ts.tv_sec = (time_t)uSeconds / 1000000;
	ts.tv_nsec = (uSeconds % 1000000) * 1000;
	
	// Sleep the requested duration.
	if (nanosleep(&ts, NULL) < 0)
		return errno;
	
	return OK;
}

// Report the model number of the device.
// Returns the model number in hex format, but it is not a hex value (e.g. Model 266 returns 0x266).
unsigned ModelNumber(s_Handle *p_HW)
{
	// Read the identification address.
	return (unsigned)((p_HW->DEV.BAR2[IDENTIFICATION] >> 32) & 0x00000FFF);
}

// Retrieve bus voltage [0] and current [1] on all power supplies and calculate power dissipation [3].
// Returns total power dissipation.
double PowerStatus(s_Handle *p_HW, s_Power *p_Power)
{
	// Read the positive bus voltage [0] and shunt voltage [1] on all three supplies and calculate power [2].
	p_Power->Supply1[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x02) >> 3) * 8.0e-3;
	p_Power->Supply1[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x01) >> 3) * 40.0e-6) / .01;
	p_Power->Supply1[2] = p_Power->Supply1[0] * p_Power->Supply1[1];
	p_Power->Supply2[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x04) >> 3) * 8.0e-3;
	p_Power->Supply2[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x03) >> 3) * 40.0e-6) / .01;
	p_Power->Supply2[2] = p_Power->Supply2[0] * p_Power->Supply2[1];
	p_Power->Supply3[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x06) >> 3) * 8.0e-3;
	p_Power->Supply3[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x05) >> 3) * 40.0e-6) / .01;
	p_Power->Supply3[2] = p_Power->Supply3[0] * p_Power->Supply3[1];

	return p_Power->Supply1[2] + p_Power->Supply2[2] + p_Power->Supply3[2];
}

// Synchronize the data serializer/deserializer clocks and frame.
// Return of zero indicates success, non-zero indicates an error condition.
int SerDesInit(s_Handle *p_HW)
{
	unsigned long long busy = 0;
	unsigned watchdog = 0;

	// Reset the watchdog counter.
	watchdog = 0;

	// Write to the clock initialize address to command the lock process to start.
	p_HW->DEV.BAR0[CLOCK_INITIALIZE] = 0x0ull;

	// Wait for the lock process to begin.
	while (!busy && (watchdog < 100))
	{
		// Read the clock initialize address to check the lock busy status.
		busy = p_HW->DEV.BAR0[CLOCK_INITIALIZE];
		// Increment the watchdog counter.
		watchdog++;
	}

	// Wait for the lock process to end.
	while (busy && (watchdog < 100))
	{
		MicrosecondTimer(10);
		// Read the clock initialize address to check the lock busy status.
		busy = p_HW->DEV.BAR0[CLOCK_INITIALIZE];
		// Increment the watchdog counter.
		watchdog++;
	}

	// Return an error if no solution was found.
	if (watchdog >= 100)
		return SERDES_NOSOL;
	else
		return OK;
}

// Program a device through a serial interface port.
// Note: The result of a serial read operation can be accessed through the ReadValue pointer.
void Serial(s_Handle *p_HW, unsigned BusSelect, unsigned MultiFunction, unsigned Instruction, unsigned InstructionSize, unsigned Data, unsigned DataSize, unsigned *p_ReadValue)
{
	unsigned long long config_payload = 0;
	unsigned max = 0xFFFFFFFF;

	// Assemble the configuration settings.
	switch (InstructionSize+DataSize)
	{
		case 8:  {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x0ull; break;}
		case 16: {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x1ull; break;}
		case 24: {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x2ull; break;}
		case 32: {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x3ull; break;}
		case 40: {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x4ull; break;}
		case 48: {config_payload = ((unsigned long long)InstructionSize/8 << 7) | ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x5ull; break;}
		default: break;
	}

	// Write to the serial configuration address.
	p_HW->DEV.BAR0[SERIAL_CONFIG] = config_payload;

	// Write to the serial payload address.
	p_HW->DEV.BAR0[SERIAL_PAYLOAD] = (unsigned long long)Instruction << DataSize | (unsigned long long)Data;

	// Initiate a serial transaction by writing to the serial command address.
	p_HW->DEV.BAR0[SERIAL_COMMAND] = 0x0ull;

	// Wait for the serial transaction to complete by monitoring the busy bit at the serial command address.
	while(p_HW->DEV.BAR0[SERIAL_COMMAND]);
	
	// Read the serial payload address to retrieve the results of a serial read operation.
	if (p_ReadValue)
		// Mask the bits to match the payload size for serial protocols other than I2C.
		*p_ReadValue = (unsigned)p_HW->DEV.BAR0[SERIAL_PAYLOAD] & (max >> (32-DataSize));
}

// Issue a software reset that is equivalent to a power-on reset.
void SoftwareReset(s_Handle *p_HW)
{
	// Write to the software reset address.
	p_HW->DEV.BAR0[SOFTWARE_RESET] = 0x0ull;
	// Pause for the reset to cross clock domains.
	MicrosecondTimer(1);
}

// Identify the device as a Sigstream product.
// Return of zero indicates SigStream product match, non-zero indicates a non-match or error condition.
int SigStreamID(s_UIOScan UIO, s_Handle *p_HW)
{
	int open = 0;
	unsigned model = 0;

	if ((open = DEVOpen(UIO, &p_HW->DEV)))
		return open;
	else
	{
		// Read the model number from the device.
		model = ModelNumber(p_HW);
		// Determine if the model number falls within the range of SigStream products (0x270 - 0x279).
		if ((model > 623) && (model < 640))
			return DEVClose(UIO, &p_HW->DEV);
		else
		{
			DEVClose(UIO, &p_HW->DEV);
			return MODEL_NUMBER;
		}
	}
}

// Issue a software trigger that is synchronous to all channels.
void SoftwareTrigger(s_Handle *p_HW)
{
	// Write to the software trigger address
	p_HW->DEV.BAR2[SOFTWARE_TRIGGER] = 0x0ull;
}

// Retrieve temperature from the device.
void TemperatureStatus(s_Handle *p_HW, s_Temperature *p_Temperature)
{
	unsigned one = 0x1, max = 0xFFFFFFFFull;
	unsigned Register[6] = {0,0,0,0,0,0};
	unsigned cnt = 0;
	
	// Configure all of the LTC2991 ports (V1-V8) for temperature measurement.
	LTC2991Write(p_HW, TEMP_MONITOR, 0x06, 0x22);
	LTC2991Write(p_HW, TEMP_MONITOR, 0x07, 0x22);
	// Perform a conversion on all of the ports and internal sensor.
	LTC2991Write(p_HW, TEMP_MONITOR, 0x01, 0xF8);
	// Wait for the conversion process to complete by monitoring the busy bit.
	do
		MicrosecondTimer(50000);
	while(LTC2991Read(p_HW, TEMP_MONITOR, 0x01) & 0x00000004);
	// Retrieve the temperature values.
	Register[0] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x0A) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x0B);
	Register[1] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x0E) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x0F);
	Register[2] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x12) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x13);
	Register[3] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x16) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x17);
	Register[4] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x1A) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x1B);
	Register[5] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x1C) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x1D);
	// Check the data valid bit on each register.
	for (cnt = 0; cnt < 6; cnt++)
		if (Register[cnt] & 0x00008000)
		{
			Register[cnt] &= 0x00001FFF;
			// Sign extend negative numbers.
			if (Register[cnt] & (one << 12)) Register[cnt] |= (max << 13);
			// Apply any necessary diode ideality correction and convert each measurement to degrees Celsius.
			switch (cnt)
			{
				case 0: {p_Temperature->Sensor[0] = ((1.004/1.010)*((Register[0]/16)+273.14))-273.15; break;}
				case 1: {p_Temperature->Sensor[1] = Register[1]/16; break;}
				case 2: {p_Temperature->Sensor[2] = Register[2]/16; break;}
				case 3: {p_Temperature->Sensor[3] = Register[3]/16; break;}
				case 4: {p_Temperature->Sensor[4] = Register[4]/16; break;}
				case 5: {p_Temperature->Vcc = 2.5+(Register[5]*315.18e-6); break;}
			}
		}
}

// Read the current time of day (TOD) seconds value.
// Note:  The fractional seconds value is not available.
// Returns the TOD seconds value.
unsigned TODSeconds(s_Handle *p_HW)
{
	// Read the TOD settings address.
	return (unsigned)(p_HW->DEV.BAR2[TOD_SETTINGS] & 0xFFFFFFFF);
}

// Set the time of day (TOD) clock and load the TSI/TSF codes.
void TODSettings(s_Handle *p_HW, s_TODSettings TOD)
{
	p_HW->DEV.BAR2[TOD_SETTINGS] =
		(unsigned long long)TOD.Seconds |
		(unsigned long long)TOD.ClockFrequency << 32;

	p_HW->DEV.BAR2[TSI_TSF_CODE] =
		(unsigned long long)TOD.TSFCode |
		(unsigned long long)TOD.TSICode << 2;
}

// Select options for the trigger inputs.
void TriggerSettings(s_Handle *p_HW, s_TriggerSettings Trigger)
{
	// Load the trigger configuration.
	p_HW->DEV.BAR2[TRIG_SETTINGS] =
		(unsigned long long)(Trigger.Source[0] & 0x3) |
		(unsigned long long)(Trigger.Source[1] & 0x3) << 2;
}
