/****************************************************  Product Functions  *****************************************************

	Red Rapids
	797 N Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to initialize the hardware specific to each SigStream model.

	Rev History:  01-20-2020:	Moved SoftwareReset() to MxxxInitialize() since it is synchronous and relies on a stable clock.
				  02-25-2019:	Inital release.

	Copyright (c) 2019  Red Rapids, Inc.

	Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
	modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software
	is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
	LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR
	IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

****************************************************  Product Functions  *****************************************************/

#include <stdlib.h>

#include "driver_functions.h"
#include "sigstream.h"
#include "global_functions.h"
#include "chip_functions.h"
#include "product_functions.h"
#include "default_hardware_config.h"

// Intitialize the Red Rapids hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int HWInitialize(s_Handle *p_HW, unsigned Model)
{
	switch (Model)
	{
		case 0x271:	// Model 271
			// Initialize hardware specific to the Model 271.
			return M271Initialize(p_HW, M271Default);
		case 0x273:	// Model 273
			// Initialize hardware specific to the Model 273.
			return M273Initialize(p_HW, M273Default);
		case 0x276:	// Model 276
			// Initialize hardware specific to the Model 276.
			return M276Initialize(p_HW, M276Default);
		case 0x277:	// Model 277
			// Initialize hardware specific to the Model 277.
			return M277Initialize(p_HW, M277Default);
		default:	// Illegal model number.
			return MODEL_NUMBER;
	}
}

// Intitialize the Model 271 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M271Initialize(s_Handle *p_HW, s_M271Devices M271Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 271 is equipped with two receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the dual ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M271_RX1TO2_ADC, M271Devices.RX1TO2_ADC.Payload[cnt][0], M271Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M271_RX1_TRIM_DAC, M271Devices.RX1_TRIM_DAC.Payload[cnt][0], M271Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M271_RX2_TRIM_DAC, M271Devices.RX2_TRIM_DAC.Payload[cnt][0], M271Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		ADL5566Write(p_HW, CMD_STAT1, M271Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M271Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M271Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	AD9512Write(p_HW, M271_CLK_DIST, 0x0000, 0x30);
	AD9512Write(p_HW, M271_CLK_DIST, 0x005A, 0x01);
	AD9512Write(p_HW, M271_CLK_DIST, 0x0000, 0x10);
	AD9512Write(p_HW, M271_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		AD9512Write(p_HW, M271_CLK_DIST, M271Devices.ClockDist.Payload[cnt][0], M271Devices.ClockDist.Payload[cnt][1]);
	
	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 273 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M273Initialize(s_Handle *p_HW, s_M273Devices M273Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;
	int status = OK;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// Query the hardware for the number of receiver and transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Reset the clock distribution chip.
	LMK01000Write(p_HW, M273_CLK_DIST, 0x0, 0x8000010);
	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the clock distribution chip input to CLKin1.
		// Note: The default input is the on-board synthesizer (CLKin0).
		M273Devices.ClockDist.Payload[9][1] &= 0xDFFFFFF;
	// Configure the clock distribution chip by loading the modified default values.
	for (cnt = 0; cnt < 10; cnt++)
		LMK01000Write(p_HW, M273_CLK_DIST, M273Devices.ClockDist.Payload[cnt][0], M273Devices.ClockDist.Payload[cnt][1]);
	// Synchronize the outputs of the clock distribution chip.
	LMK01000Sync(p_HW, M273_CLK_DIST_SYNC);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Reset the dual ADC that is assigned to RX channels #1 and #2.
	AD9652Write(p_HW, M273_RX1TO2_ADC, 0x0, 0x3C);
	// Wait 500 us for the SPI interface to become available after reset.
	MicrosecondTimer(500);
	// Configure the dual ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 26; cnt++)
		AD9652Write(p_HW, M273_RX1TO2_ADC, M273Devices.RX1TO2_ADC.Payload[cnt][0], M273Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the RX trim DACs from the default structure values.
	// Note: Trim DACs are only present on RX DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M273_RX1_TRIM_DAC, M273Devices.RX1_TRIM_DAC.Payload[cnt][0], M273Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M273_RX2_TRIM_DAC, M273Devices.RX2_TRIM_DAC.Payload[cnt][0], M273Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the RX amplifier enables from the default structure values.
	// Note: Amplifiers are only present on RX DC coupled units.
	ADA4927Write(p_HW, CMD_STAT1, M273Devices.RX1_AMP_ENBL.Enable[0], M273Devices.RX1_AMP_ENBL.Enable[1]);
	ADA4927Write(p_HW, CMD_STAT1, M273Devices.RX2_AMP_ENBL.Enable[0], M273Devices.RX2_AMP_ENBL.Enable[1]);
	
	// Phase align the clocks to valid data.
	if ((status = ClockInitLMK01000(p_HW, M273Default.ClockDist.Payload[0][1], M273Default.ClockDist.Payload[4][1], M273Default.ClockDist.Payload[5][1])))
		return status;
	
	// Initiate the ADC fast start-up calibration.
	AD9652Cal(p_HW, M273_RX1TO2_ADC);
	
	return OK;
}

// Intitialize the Model 276 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M276Initialize(s_Handle *p_HW, s_M276Devices M276Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The standard Model 276 has two receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the ADC from the default structure values.
	for (cnt = 0; cnt < 11; cnt++)
		ADC12D1600Write(p_HW, M276_RX1TO2_ADC, M276Devices.RX1TO2_ADC.Payload[cnt][0], M276Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M276_RX1_TRIM_DAC, M276Devices.RX1_TRIM_DAC.Payload[cnt][0], M276Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M276_RX2_TRIM_DAC, M276Devices.RX2_TRIM_DAC.Payload[cnt][0], M276Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		ADL5566Write(p_HW, CMD_STAT1, M276Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M276Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);
	
	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}

// Intitialize the Model 277 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
int M277Initialize(s_Handle *p_HW, s_M277Devices M277Devices)
{
	unsigned cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 277 can is equipped with four receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);

	// Configure the I/O signal direction and impedance of the command/status port expander.
	// Note: The 100 MHz configuration flash clock is disabled.
	FXL6408Write(p_HW, CMD_STAT1, 0x03, 0xF8);
	FXL6408Write(p_HW, CMD_STAT1, 0x07, 0x07);
	// Mask interrupts from the GPIO port expander.
	FXL6408Write(p_HW, CMD_STAT1, 0x11, 0xFF);

	// Configure the first ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M277_RX1TO2_ADC, M277Devices.RX1TO2_ADC.Payload[cnt][0], M277Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the second ADC that is assigned to RX channels #3 and #4 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		ADS42LB69Write(p_HW, M277_RX3TO4_ADC, M277Devices.RX3TO4_ADC.Payload[cnt][0], M277Devices.RX3TO4_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		LTC1661Write(p_HW, M277_RX1_TRIM_DAC, M277Devices.RX1_TRIM_DAC.Payload[cnt][0], M277Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M277_RX2_TRIM_DAC, M277Devices.RX2_TRIM_DAC.Payload[cnt][0], M277Devices.RX2_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M277_RX3_TRIM_DAC, M277Devices.RX3_TRIM_DAC.Payload[cnt][0], M277Devices.RX3_TRIM_DAC.Payload[cnt][1]);
		LTC1661Write(p_HW, M277_RX4_TRIM_DAC, M277Devices.RX4_TRIM_DAC.Payload[cnt][0], M277Devices.RX4_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
	{
		ADL5566Write(p_HW, CMD_STAT1, M277Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M277Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);
		ADL5566Write(p_HW, CMD_STAT1, M277Devices.RX3TO4_AMP_ENBL.Enable[cnt][0], M277Devices.RX3TO4_AMP_ENBL.Enable[cnt][1]);
	}

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M277Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	AD9512Write(p_HW, M277_CLK_DIST, 0x0000, 0x30);
	AD9512Write(p_HW, M277_CLK_DIST, 0x005A, 0x01);
	AD9512Write(p_HW, M277_CLK_DIST, 0x0000, 0x10);
	AD9512Write(p_HW, M277_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		AD9512Write(p_HW, M277_CLK_DIST, M277Devices.ClockDist.Payload[cnt][0], M277Devices.ClockDist.Payload[cnt][1]);

	// Software reset the device now that the sample clock is stable.
	SoftwareReset(p_HW);

	// Phase align the clocks to valid data.
	return ClockInit(p_HW);
}
